"""supporting functions"""

# AUTOGENERATED! DO NOT EDIT! File to edit: ../../nbs/utils/files.ipynb.

# %% auto 0
__all__ = ['rename_filepath_to_match_datatype', 'change_file_suffix', 'detect_encoding', 'read_html_file_as_soup',
           'upsert_folder', 'upsert_file', 'get_all_files_and_folders', 'update_env', 'output_content_and_update_env',
           'ImageBlockConverter', 'md', 'convert_html_to_markdown', 'convert_remove_html_tags', 'download_zip',
           'download_pptx', 'download_markdown', 'generate_filename_from_url']

# %% ../../nbs/utils/files.ipynb 3
from typing import List, Tuple, Union, Callable, Any

import os
import json
import chardet
import dotenv

import pathlib
import shutil

from bs4 import BeautifulSoup
from markdownify import MarkdownConverter
import PIL

import datetime as dt


import pptx2md

import zipfile
import io

from urllib.parse import urlparse

# %% ../../nbs/utils/files.ipynb 7
def rename_filepath_to_match_datatype(data, file_path):
    is_path_ext = os.path.splitext(file_path)[-1].lower()

    old_suffix = pathlib.Path(file_path).suffix if is_path_ext else None

    new_suffix = ""

    if isinstance(data, str) or isinstance(data, bytes) or isinstance(data, bytearray):
        new_suffix = ".txt"
    if isinstance(data, dict):
        new_suffix = ".json"

    file_path = file_path + new_suffix

    if old_suffix:
        file_path = file_path.replace(old_suffix, "")

    return file_path

# %% ../../nbs/utils/files.ipynb 9
def change_file_suffix(filename, new_extension, change_if_exists: bool = True):
    is_exist = os.path.exists(filename)

    p = pathlib.PurePath(filename)
    p = p.with_suffix(new_extension)

    if is_exist and change_if_exists:
        p = pathlib.Path(filename)
        p = p.rename(p.with_suffix(new_extension))

    return str(p)

# %% ../../nbs/utils/files.ipynb 11
def detect_encoding(file_path, debug_prn: bool = False):
    detector = chardet.universaldetector.UniversalDetector()
    with open(file_path, "rb") as f:
        for line in f:
            detector.feed(line)
            if detector.done:
                break
    detector.close()

    encoding = detector.result

    return encoding

# %% ../../nbs/utils/files.ipynb 13
def read_html_file_as_soup(
    file_path, is_convert_to_soup: bool = True
) -> Union[str, BeautifulSoup]:

    if not os.path.exists(file_path):
        raise FileNotFoundError(file_path)

    page_encoding = detect_encoding(file_path)

    with open(file_path, encoding=page_encoding["encoding"]) as fp:
        if is_convert_to_soup:
            return BeautifulSoup(fp, "lxml")

        return fp.read()

# %% ../../nbs/utils/files.ipynb 16
def upsert_folder(folder_path: str, debug_prn: bool = False, replace_folder=False):

    folder_path = os.path.dirname(folder_path)

    if replace_folder and os.path.exists(folder_path) and os.path.isdir(folder_path):
        shutil.rmtree(folder_path)

    if debug_prn:
        print(
            {
                "upsert_folder": os.path.abspath(folder_path),
                "is_exist": os.path.exists(folder_path),
            }
        )

    if not os.path.exists(folder_path):
        os.makedirs(folder_path)

# %% ../../nbs/utils/files.ipynb 17
def upsert_file(
    file_path: str,
    content: Any = None,
    write_fn: Callable = None,
    file_update_method="w",
    encoding="utf-8",
    debug_prn: bool = False,
    replace_folder: bool = False,
) -> bool:

    upsert_folder(
        folder_path=file_path, debug_prn=debug_prn, replace_folder=replace_folder
    )

    with open(file_path, file_update_method, encoding=encoding) as file:
        if not write_fn:
            file.write(content or "")
            return True

        return write_fn(file=file, content=content)

# %% ../../nbs/utils/files.ipynb 18
def get_all_files_and_folders(
    directory, file_type=None  # to only retrieve a specific file type
) -> Union[Tuple, List]:
    """
    walk a directory and retrieve a list of files and a list of directory
    returns Tuple of (file_ls , dir_ls) OR file_ls if file_type supplied
    """
    if not os.path.exists(directory):
        raise FileNotFoundError(directory)

    file_ls = []
    dir_ls = []
    for root, dirs, files in os.walk(directory):
        for name in files:
            if file_type:
                if not name.lower().endswith(file_type.lower()):
                    continue
            file_ls.append(os.path.join(root, name))

        if file_type:
            continue

        for name in dirs:
            dir_ls.append(os.path.join(root, name))

    if file_type:
        return file_ls

    return file_ls, dir_ls

# %% ../../nbs/utils/files.ipynb 21
def update_env(
    env_path: str, env_key: str, content: str, debug_prn: bool = False
) -> dict:
    """
    updates a .env file with a key value pair
    then reloads the env_file
    """

    if not os.path.exists(env_path):
        with open(env_path, "w", encoding="utf-8") as f:
            f.write("")

    quote_mode = "always"

    if isinstance(content, dict):
        quote_mode = "never"
        content = json.dumps(content)

    if debug_prn:
        from pprint import pprint

        pprint(
            {
                "env_path": env_path,
                "key": env_key,
                "value": content,
                "type": type(content),
                "quote_mode": quote_mode,
            }
        )

    dotenv.set_key(env_path, env_key, content, quote_mode=quote_mode)

    dotenv.set_key(env_path, "env_last_modified", f"updated - {dt.date.today()}")

    dotenv.load_dotenv(env_path, override=True)

    return {env_key: os.getenv(env_key)}

# %% ../../nbs/utils/files.ipynb 22
def output_content_and_update_env(
    content, output_path=None, env_path=None, env_key=None, debug_prn: bool = False
):
    if env_key:
        os.environ[env_key] = content

    if output_path:
        upsert_file(file_path=output_path, content=content, debug_prn=debug_prn)

    if env_path:
        update_env(
            env_path=env_path, env_key=env_key, content=content, debug_prn=debug_prn
        )

# %% ../../nbs/utils/files.ipynb 24
class ImageBlockConverter(MarkdownConverter):
    """
    Create a custom MarkdownConverter that adds two newlines after an image
    """

    def convert_img(self, el, text, convert_as_inline, is_resize: bool = True):
        """
        custom image downloader for ImabeBlockConverter
        will handle resize
        """

        if is_resize:
            style_obj = {
                (obj.split(":")[0].strip()): obj.split(":")[1].strip()
                for obj in el.get("style").split(";")
                if ":" in obj
            }

            file_path = os.path.join(
                os.path.dirname(self.options["file_path"]), el["src"]
            )

            image = PIL.Image.open(file_path)

            width = style_obj["width"].replace("px", "")
            width = int(float(width))

            height = style_obj["height"].replace("px", "")
            height = int(float(height))

            new_image = image.resize((width, height))
            new_image.save(file_path)

        return super().convert_img(el, text, convert_as_inline)


def md(html, **options):
    """Create shorthand method for handling conversion"""
    return ImageBlockConverter(**options).convert(html)

# %% ../../nbs/utils/files.ipynb 25
def convert_html_to_markdown(file_path):
    """converts html file to markdown in place"""

    with open(file_path, encoding="utf-8") as f:
        html = f.read()

    markdown_content = md(
        str(html),
        keep_inline_images_in=["td", "span"],
        file_path=file_path,
        is_resize=True,
    )

    md_path = file_path.replace(".html", ".md")

    with open(md_path, "w+", encoding="utf-8") as f:
        f.write(markdown_content)

    return

# %% ../../nbs/utils/files.ipynb 26
def convert_remove_html_tags(html_string: str) -> str:
    """
    Clean an HTML string using BeautifulSoup.

    Args:
        html_string (str): The HTML string.

    Returns:
        str: The cleaned string.
    """
    soup = BeautifulSoup(html_string, "html.parser")
    clean_string = soup.get_text()
    return clean_string

# %% ../../nbs/utils/files.ipynb 28
def download_zip(
    content,
    output_folder,
    is_convert_to_markdown: bool = True,
    replace_folder=True,
    **kwargs,
):
    """save bytes content to a zip file then convert html to markdown"""

    upsert_folder(folder_path=output_folder, replace_folder=replace_folder)

    zip = zipfile.ZipFile(io.BytesIO(content), "r")
    zip.extractall(output_folder)

    file_ls = os.listdir(output_folder)

    # rename the html file to index.html
    for file_name in file_ls:
        if file_name.endswith(".html"):
            output_index = os.path.join(output_folder, "index.html")
            os.replace(os.path.join(output_folder, file_name), output_index)

            if is_convert_to_markdown:
                convert_html_to_markdown(os.path.join(output_folder, "index.html"))

    return f"successfully downloaded zip to {output_folder}"

# %% ../../nbs/utils/files.ipynb 31
def download_pptx(
    content,
    output_folder,
    is_convert_to_markdown: bool = True,
    replace_folder: bool = True,
    **kwargs,
):
    """save bytes content to a pptx file then converts to markdown"""

    output_ppt_index = os.path.join(output_folder, "index.pptx")

    upsert_file(output_ppt_index, replace_folder=replace_folder)

    with open(output_ppt_index, "wb+") as binary_file:
        # Write bytes to file
        binary_file.write(content)

    if is_convert_to_markdown:
        pptx2md.convert(
            pptx2md.ConversionConfig(
                pptx_path=output_ppt_index,
                output_path=os.path.join(output_folder, "index.md"),
                image_dir=os.path.join(output_folder, "images"),
            )
        )

    return f"successfully downloaded content to {output_folder}"

# %% ../../nbs/utils/files.ipynb 34
def download_markdown(content, output_folder, replace_folder: bool = True):

    upsert_file(
        file_path=output_folder, content=md(content), replace_folder=replace_folder
    )

# %% ../../nbs/utils/files.ipynb 36
def generate_filename_from_url(url, download_folder=None, file_name=None) -> str:
    parsed_url = urlparse(url)

    file_path = "_".join([str for str in parsed_url[2].split("/") if str])

    if download_folder:
        file_path = os.path.join(download_folder, file_path)

    if file_name:
        file_path = os.path.join(file_path, file_name)

    return file_path



# AUTOGENERATED! DO NOT EDIT! File to edit: ../../nbs/scraper/routes.ipynb.

# %% auto 0
__all__ = ['DEFAULT_LINUX_PATH', 'generate_webdriver', 'WebDriverException', 'WebDriverException_Timeout', 'wait_and_return',
           'get_pagesource', 'authenticate_domo_driver']

# %% ../../nbs/scraper/routes.ipynb 2
from selenium.webdriver.common.by import By

# %% ../../nbs/scraper/routes.ipynb 3
import os

from selenium import webdriver

from selenium.webdriver.support.ui import WebDriverWait
from selenium.webdriver.support import expected_conditions as EC
from selenium.common.exceptions import TimeoutException


from bs4 import BeautifulSoup
import time

# %% ../../nbs/scraper/routes.ipynb 9
DEFAULT_LINUX_PATH = "/usr//bin/chromedriver"

# def save_cookie(driver, cookie_path = './COOKIES'):
#     deut.upsert_folder(cookie_path)

#     with open(cookie_path, 'wb') as cookie_file:
#         pickle.dump(driver.get_cookies(), cookie_file)


# def load_cookie(driver, cookie_path):

#      if not os.path.exists(cookie_path):
#          return False

#      with open(cookie_path, 'rb') as cookie_file:
#          cookies = pickle.load(cookie_file)

#          for cookie in cookies:
#              driver.add_cookie(cookie)


def generate_webdriver(
    driver_path=None,
    is_headless: bool = True,
) -> webdriver:

    options = webdriver.ChromeOptions()
    # run Selenium in headless mode

    if is_headless:
        options.add_argument("--headless")

    options.add_argument("--no-sandbox")  # needed for LINUX environments

    if driver_path:
        options.add_argument(driver_path)

    driver = webdriver.Chrome(options=options)

    return driver

# %% ../../nbs/scraper/routes.ipynb 10
class WebDriverException(Exception):
    url: str
    message: str

    def __init__(self, url, message):
        super.__init__(f"WebdriverException: {message} while retrieving {url}")


class WebDriverException_Timeout(Exception):
    def __init__(self, url, max_sleep_time):
        super().__init__(
            f"WebDriverException_Timeout: {url} did not load within {max_sleep_time} seconds."
        )


def wait_and_return(
    driver: webdriver,
    search_criteria_tuple: tuple,
    max_sleep_time=15,
    return_soup: bool = False,
):
    """
    Wait for the element to be present and return it.
    """

    try:
        ele = WebDriverWait(driver, timeout=max_sleep_time).until(
            EC.presence_of_element_located(search_criteria_tuple)
        )

        if return_soup:
            return BeautifulSoup(ele.get_attribute("outerHTML"), "lxml")
        return ele

    except TimeoutException as e:
        print(e)
        raise WebDriverException_Timeout(driver.current_url, max_sleep_time)

    except Exception as e:
        raise WebDriverException(url=driver.current_url, message=e)


def get_pagesource(
    url: str = None,
    driver: webdriver = None,
    search_criteria_tuple=None,  # for dynamically rendered pages, pass a WebDriverWait search tuple (search_element, element name)
    max_sleep_time=15,
    return_soup: bool = False,
    return_entire_page: bool = True,  # to return entire page instead of just search_criteria
    debug_prn: bool = False,
):
    """retrieve page_source, for async loaded content, can add search_criteria."""
    is_close_driver = False if driver else True

    driver = driver or generate_webdriver()

    try:
        if debug_prn:
            message = f"💤 pagesource: retrieving {url} with max sleep time {max_sleep_time} 💤"
            if search_criteria_tuple:
                message += f"using {str(search_criteria_tuple)}"

            print(message)

        if url:
            driver.get(url)

        if search_criteria_tuple:
            ele = wait_and_return(
                driver=driver,
                max_sleep_time=max_sleep_time,
                search_criteria_tuple=search_criteria_tuple,
                return_soup=return_soup,
            )

            if not return_entire_page:
                return ele

        if return_soup:
            return BeautifulSoup(driver.page_source, "lxml")

        return driver.page_source

    finally:
        if is_close_driver:
            driver.quit()

# %% ../../nbs/scraper/routes.ipynb 14
def authenticate_domo_driver(
    driver: webdriver,
    domo_instance,
    domo_username,
    domo_password,
    sleep_time=4,  # estimate of time to load pages, increase for slower internet connections
) -> webdriver:
    """returns a driver with (non sso) username and password authentication to Domo assets"""

    # authenticate against your instance
    url = f"https://{domo_instance}.domo.com/auth/index"
    driver.get(url)

    button = wait_and_return(
        driver=driver,
        search_criteria_tuple=(By.CLASS_NAME, "login-btn"),
        max_sleep_time=15,
        return_soup=False,
    )

    form_username = wait_and_return(
        driver=driver, search_criteria_tuple=(By.NAME, "username"), return_soup=False
    )

    form_password = wait_and_return(
        driver=driver, search_criteria_tuple=(By.NAME, "password"), return_soup=False
    )

    # Sending input
    form_username.clear()
    form_username.send_keys(domo_username)
    form_password.clear()
    form_password.send_keys(domo_password)
    button.click()

    time.sleep(sleep_time)

    # authenticate to domo resources
    url = f"https://domo-support.domo.com/s/"
    driver.get(url)

    domain_input = wait_and_return(
        driver=driver,
        search_criteria_tuple=(By.CLASS_NAME, "domain-input"),
        return_soup=False,
    )
    button = wait_and_return(
        driver=driver,
        search_criteria_tuple=(By.CLASS_NAME, "continue-button"),
        return_soup=False,
    )

    domain_input.send_keys(domo_instance)
    button.click()

    time.sleep(sleep_time)

    return driver



"""supporting functions"""

# AUTOGENERATED! DO NOT EDIT! File to edit: ../../nbs/utils/convert.ipynb.

# %% auto 0
__all__ = ['convert_json_to_obj', 'convert_enum_member_to_name', 'convert_str_to_snake_case', 'convert_str_remove_accents',
           'convert_str_keep_alphanumeric', 'convert_str_remove_emojis', 'convert_str_file_name', 'convert_str_to_date',
           'convert_datetime_to_str', 'sort_dictionary']

# %% ../../nbs/utils/convert.ipynb 3
from enum import Enum
import json
from typing import Union
import re
import unicodedata

import datetime as dt
from dateutil.parser import parse as dtu_parse

# %% ../../nbs/utils/convert.ipynb 5
def convert_json_to_obj(json_str: Union[str, dict]) -> dict:
    """
    receives a json string and converts to a dict
    """
    if isinstance(json_str, str):
        return json.loads(json_str)
    return json_str

# %% ../../nbs/utils/convert.ipynb 6
def convert_enum_member_to_name(enum_member):
    return enum_member.name if isinstance(enum_member, Enum) else enum_member

# %% ../../nbs/utils/convert.ipynb 7
def convert_str_to_snake_case(text_str):
    """converts 'snake_case_str' to 'snakeCaseStr'"""

    return text_str.replace(" ", "_").lower()

# %% ../../nbs/utils/convert.ipynb 8
def convert_str_remove_accents(text_str: str) -> str:
    return "".join(
        c
        for c in unicodedata.normalize("NFD", text_str)
        if unicodedata.category(c) != "Mn"
    )

# %% ../../nbs/utils/convert.ipynb 10
def convert_str_keep_alphanumeric(text_str) -> str:
    pattern = r"[^0-9a-zA-Z_\-\s]+"

    return re.sub(pattern, "", text_str)

# %% ../../nbs/utils/convert.ipynb 11
def convert_str_remove_emojis(text_str: str) -> str:
    emoj = re.compile(
        "["
        "\U0001F600-\U0001F64F"  # emoticons
        "\U0001F300-\U0001F5FF"  # symbols & pictographs
        "\U0001F680-\U0001F6FF"  # transport & map symbols
        "\U0001F1E0-\U0001F1FF"  # flags (iOS)
        "\U00002500-\U00002BEF"  # chinese char
        "\U00002702-\U000027B0"
        "\U00002702-\U000027B0"
        "\U000024C2-\U0001F251"
        "\U0001f926-\U0001f937"
        "\U00010000-\U0010ffff"
        "\u2640-\u2642"
        "\u2600-\u2B55"
        "\u200d"
        "\u23cf"
        "\u23e9"
        "\u231a"
        "\ufe0f"  # dingbats
        "\u3030"
        "]+",
        re.UNICODE,
    )
    return re.sub(emoj, "", text_str)

# %% ../../nbs/utils/convert.ipynb 13
def convert_str_file_name(text_str: str) -> str:
    """convert strings to clean file name or url"""

    return convert_str_keep_alphanumeric(
        convert_str_to_snake_case(convert_str_remove_accents(text_str))
    )

# %% ../../nbs/utils/convert.ipynb 15
def convert_str_to_date(datefield: str) -> dt.datetime:
    """converts string date to datetime object"""
    return dtu_parse(datefield) if datefield else None

# %% ../../nbs/utils/convert.ipynb 17
def convert_datetime_to_str(date_date: dt.datetime, format_str="%Y-%m-%d %H:%M") -> str:
    return date_date.strftime(format_str) if date_date else None

# %% ../../nbs/utils/convert.ipynb 18
def sort_dictionary(my_dict):
    my_keys = list(my_dict.keys())
    my_keys.sort()
    return {key: my_dict[key] for key in my_keys}



"""base classes and functions for this library"""

# AUTOGENERATED! DO NOT EDIT! File to edit: ../nbs/client/client.ipynb.

# %% ../nbs/client/client.ipynb 3
from __future__ import annotations

import os
from typing import Any, Union
from dataclasses import dataclass, field
from abc import abstractmethod, ABC

from urllib.parse import urlparse

import httpx
import json

from pprint import pprint
import domolibrary_extensions.utils.utils as ut

# %% auto 0
__all__ = ['looper_offset_params', 'Auth', 'ResponseGetData', 'get_cache', 'update_cache', 'BaseError_Validation', 'BaseError',
           'get_data', 'get_data_stream', 'looper']

# %% ../nbs/client/client.ipynb 5
@dataclass
class Auth(ABC):
    """Base class for authentication"""

    @abstractmethod
    def generate_auth_header(self) -> dict:
        """Get the headers for the authentication"""
        pass

# %% ../nbs/client/client.ipynb 6
@dataclass
class ResponseGetData:
    """class for returning data from any route"""

    is_from_cache: bool
    is_success: bool

    status: int
    response: Any
    auth: Any = field(repr=False, default=None)

    def __post_init__(self):
        self.is_success = True if self.status >= 200 and self.status <= 399 else False

    @classmethod
    def _from_httpx(cls, res: httpx.Response, auth: Any = None):
        data = None
        try:
            data = res.json()

        except:
            data = res.text

        return cls(
            status=res.status_code,
            response=data,
            is_success=res.is_success,
            is_from_cache=False,
            auth=auth,
        )

    @classmethod
    def _from_stream(
        cls,
        res: httpx.Response,
        content,
        auth: Any = None,
    ):
        if not res.is_success:
            content = res.json()

        return cls(
            status=res.status_code,
            response=content,
            is_success=res.is_success,
            is_from_cache=False,
            auth=auth,
        )

    @classmethod
    def _from_cache(cls, data: dict = None, auth: Any = None):
        return cls(
            status=200,
            response=data,
            is_success=True,
            is_from_cache=True,
            auth=auth,
        )

# %% ../nbs/client/client.ipynb 11
def get_cache(cache_path: str, debug_prn: bool = False) -> Union[dict, None]:
    """function for getting cached data from json file"""

    json_data = None
    ut.upsert_folder(folder_path=cache_path, debug_prn=debug_prn)

    try:
        with open(cache_path, "r", encoding="utf-8") as file:
            json_data = json.load(file)

    except (FileNotFoundError, json.JSONDecodeError) as e:
        with open(cache_path, "w+", encoding="utf-8") as file:
            pass
        json_data = None

    if json_data:
        if debug_prn:
            print(f"🚀 Using cached data in {cache_path}")

    return json_data


def update_cache(cache_path: str, data: Any, debug_prn: bool = False):
    ut.upsert_folder(cache_path)

    cache_path = ut.rename_filepath_to_match_datatype(data, cache_path)

    if debug_prn:
        print(f"updating {type(data)} content to {cache_path}")

    if isinstance(data, bytearray) or isinstance(data, bytes):
        with open(cache_path, "wb") as bf:
            return bf.write(data)

    with open(cache_path, "w+", encoding="utf-8") as fp:
        if isinstance(data, dict):
            return json.dump(data, fp)

        if isinstance(data, str):
            return fp.write(data)

# %% ../nbs/client/client.ipynb 14
class BaseError_Validation(Exception):
    def __init__(self, message):
        super().__init__(message)


class BaseError(Exception):
    def __init__(
        self, instance=None, entity_id=None, message=None, res: ResponseGetData = None
    ):
        if not (instance and message) and not res:
            raise BaseError_Validation(
                "must include instance and message or ResponseGetData class"
            )

        message = message or res.response.get("Message", res.response)

        if entity_id:
            message = f"{message} for {entity_id}"

        if res:
            message = (
                f"status {res.status} || {message} in {instance or res.auth.instance}"
            )
        super().__init__(message)

# %% ../nbs/client/client.ipynb 16
def prepare_fetch(
    url: str,
    params: dict = None,
    auth: Auth = None,
    headers: dict = None,
    body: dict = None,
):
    """base function to prepare a fetch operation"""

    headers = headers or {"Accept": "application/json"}

    if auth:
        headers = {**headers, **auth.generate_auth_header()}

    return headers, url, params, body

# %% ../nbs/client/client.ipynb 17
def _generate_cache_name(url):
    uparse = urlparse(url)

    ufolder = "".join([uparse.netloc.replace(".", "_")])
    upath = uparse.path.replace(".", "_").split("/")
    uname = f"_{upath.pop(-1)}.json"

    return f"./CACHE/{ufolder}{'/'.join(upath)}/{uname}"

# %% ../nbs/client/client.ipynb 19
async def get_data(
    url: str,
    method: str,
    cache_path: str = None,
    is_ignore_cache: bool = False,
    headers: dict = None,
    params: dict = None,
    body=None,
    auth: Auth = None,
    parent_class: str = None,
    debug_api: bool = False,
    debug_prn: bool = False,
    client: httpx.AsyncClient = None,
    is_verify_ssl: bool = False,
    timeout: int = 10,
) -> ResponseGetData:
    """wrapper for httpx Request library, always use with jiralibrary class"""

    cache_path = cache_path or _generate_cache_name(url)

    if not is_ignore_cache and cache_path:
        json_data = get_cache(cache_path=cache_path, debug_prn=debug_prn)

        if json_data:
            return ResponseGetData._from_cache(data=json_data, auth=auth)

    is_close_session = False if client else True
    client = client or httpx.AsyncClient(verify=is_verify_ssl, follow_redirects=True)

    headers, url, params, body = prepare_fetch(
        url=url,
        params=params,
        auth=auth,
        headers=headers,
        body=body,
    )

    if debug_api:
        pprint(
            {
                "method": method,
                "headers": headers,
                "url": url,
                "params": params,
                "body": body,
                "cache_file_path": cache_path,
                "debug_api": debug_api,
                "parent_class": parent_class,
            }
        )

    res = None
    if method.upper() == "GET":
        res = await client.get(
            url=url,
            headers=headers,
            params=params,
            follow_redirects=True,
            timeout=timeout,
        )
    else:
        if headers.get("Content-Type") == "application/json":
            res = await getattr(client, method.lower())(
                url=url, headers=headers, params=params, json=body, timeout=timeout
            )
        else:
            res = await getattr(client, method.lower())(
                url=url, headers=headers, params=params, data=body, timeout=timeout
            )

    if is_close_session:
        await client.aclose()

    rgd = ResponseGetData._from_httpx(res, auth=auth)

    if rgd.is_success:
        update_cache(cache_path=cache_path, data=rgd.response, debug_prn=debug_prn)

    return rgd

# %% ../nbs/client/client.ipynb 21
async def get_data_stream(
    url: str,
    cache_path: str = None,
    is_ignore_cache: bool = False,
    headers: dict = None,
    params: dict = None,
    body=None,
    auth: Auth = None,
    parent_class: str = None,
    debug_api: bool = False,
    debug_prn: bool = False,
    client: httpx.AsyncClient = None,
    is_text: bool = False,  # if false will interpret as bytes
    is_verify_ssl: bool = False,
) -> ResponseGetData:
    """wrapper for httpx Request library, always use with jiralibrary class"""

    cache_path = cache_path or _generate_cache_name(url)

    if not is_ignore_cache and cache_path:
        json_data = get_cache(cache_path=cache_path, debug_prn=debug_prn)

        if json_data:
            return ResponseGetData._from_cache(data=json_data, auth=auth)

    is_close_session = False if client else True

    client = client or httpx.AsyncClient(verify=is_verify_ssl)

    headers, url, params, body = prepare_fetch(
        url=url,
        params=params,
        auth=auth,
        headers=headers,
        body=body,
    )

    if debug_api:
        pprint(
            {
                "headers": headers,
                "url": url,
                "params": params,
                "body": body,
                "cache_file_path": cache_path,
                "debug_api": debug_api,
                "parent_class": parent_class,
            }
        )

    content = bytearray()

    async with client.stream(
        method="GET",
        url=url,
        headers=headers,
        params=params,
        follow_redirects=True,
    ) as res:
        if is_text:
            async for chunk in res.aiter_text():
                content += chunk

        else:
            async for chunk in res.aiter_bytes():
                content += chunk

    if is_close_session:
        await client.aclose()

    rgd = ResponseGetData._from_stream(res, auth=auth, content=content)

    if rgd.is_success:
        update_cache(cache_path=cache_path, data=rgd.response, debug_prn=debug_prn)

    return rgd

# %% ../nbs/client/client.ipynb 24
looper_offset_params = {"offset": "offset", "limit": "limit"}


async def looper(
    url,
    client: httpx.AsyncClient,
    auth: Auth,
    arr_fn,
    offset=0,
    limit=50,
    params: dict = None,
    body: dict = None,
    offset_params: dict = None,
    offset_params_is_header: bool = False,
    debug_loop: bool = False,
    debug_api: bool = False,
    debug_prn: bool = False,
    method="GET",
    is_verify_ssl: bool = False,
    is_ignore_cache: bool = False,
    cache_path: str = None,
    return_raw: bool = False,  # will break the looper after the first request and ignore the array processing step.
    timeout: int = 10,
    **kwargs
):
    offset_params = offset_params or looper_offset_params

    cache_path = cache_path or _generate_cache_name(url)

    if not is_ignore_cache and cache_path:
        json_data = get_cache(cache_path=cache_path, debug_prn=debug_prn)

        if json_data:
            return ResponseGetData._from_cache(data=json_data, auth=auth)

    final_array = []
    keep_looping = True

    while keep_looping:
        new_params = params.copy() if params else {}
        new_body = body.copy() if body else {}
        new_offset = {
            offset_params["offset"]: offset,
            offset_params["limit"]: limit,
        }
        if offset_params_is_header:
            new_params = {**new_params, **new_offset}
        else:
            new_body = {**new_body, **new_offset}

        res = await get_data(
            is_ignore_cache=True,
            auth=auth,
            url=url,
            method=method,
            params=new_params,
            debug_api=debug_api,
            debug_prn=debug_prn,
            client=client,
            is_verify_ssl=is_verify_ssl,
            body=new_body,
            timeout=timeout,
            **kwargs
        )

        if not res.is_success or return_raw:
            return res

        new_array = arr_fn(res)

        if debug_loop:
            print(new_params, new_body, new_array)

        if not new_array or len(new_array) == 0:
            keep_looping = False

        final_array += new_array
        offset += limit

    res.response = final_array

    if res.is_success:
        update_cache(cache_path=cache_path, data=res.response, debug_prn=debug_prn)

    return res



# AUTOGENERATED! DO NOT EDIT! File to edit: ../../nbs/google/GDrive.ipynb.

# %% auto 0
__all__ = ['DEFAULT_GDRIVE_FIELDS', 'DEFAULT_GFOLDER_FIELDS', 'get_file_by_id', 'get_folder_contents_by_id',
           'walk_folder_and_flatten', 'walk_folder', 'GDrive_MimeType_Enum', 'GDrive_ExportError', 'export_file_as',
           'export_as_pptx', 'export_as_zip', 'ExportAs_Enum', 'GDrive_ServiceRequired', 'GDrive_File', 'GDrive_Folder']

# %% ../../nbs/google/GDrive.ipynb 2
from functools import partial

from dataclasses import dataclass, field
from enum import Enum
from typing import List, Any, Optional, Callable

import datetime as dt

from googleapiclient.errors import HttpError

from domolibrary_extensions.utils.utils import (
    upsert_folder,
    convert_str_to_date,
    download_zip,
    download_pptx,
)

import domolibrary_extensions.utils.execution as ce
import domolibrary_extensions.google.GAuth as ga

# %% ../../nbs/google/GDrive.ipynb 5
DEFAULT_GDRIVE_FIELDS = [
    "id",
    "webViewLink",
    "name",
    "mimeType",
    " modifiedTime",
    "parents",
]


async def get_file_by_id(
    service: ga.Resource,
    file_id: str,
    fields: List[
        str
    ] = None,  # fields to retrieve from the GDrive API (see Google Documentation for additional options)
    parent_ls: List[str] = None,
):
    parent_ls = parent_ls or []

    fields = fields or DEFAULT_GDRIVE_FIELDS

    file = service.files().get(fileId=file_id, fields=", ".join(fields)).execute()

    parent_ls += [
        parent for parent in file.get("parents", []) if parent not in parent_ls
    ]

    file.update({"parent_ls": parent_ls})

    return file

# %% ../../nbs/google/GDrive.ipynb 7
DEFAULT_GFOLDER_FIELDS = [
    "nextPageToken",
    "files(id,webViewLink, name, mimeType,modifiedTime, parents)",
]


async def get_folder_contents_by_id(
    service: ga.Resource,
    folder_id: str,
    fields: List[str] = None,
    parent_ls: List[str] = None,  # used by recursive query to build path to file
    just_folders: bool = False,
    return_raw: bool = False,
):

    fields = fields or DEFAULT_GFOLDER_FIELDS

    res = None
    file_ls = []
    page_token = None

    query = f"'{folder_id}' in parents"

    if just_folders:

        query += " and mimeType = 'application/vnd.google-apps.folder'"

    while True:

        res = (
            service.files()
            .list(
                q=query,
                pageSize=10,
                fields=",".join(fields),
                pageToken=page_token,
            )
            .execute()
        )

        file_ls.extend(res.get("files", []))

        page_token = res.get("nextPageToken")

        if not page_token:
            break

    if return_raw:
        res["files"] = file_ls
        return res

    parent_ls = parent_ls.copy() if parent_ls else []

    parent_ls = list(set(parent_ls + [folder_id]))

    for file in file_ls:

        file.update(
            {
                "parent_ls": parent_ls,
                "is_folder": file.get("mimeType")
                == "application/vnd.google-apps.folder",
            }
        )

    return file_ls

# %% ../../nbs/google/GDrive.ipynb 9
async def walk_folder_and_flatten(
    service: ga.Resource,
    folder_id: str,
    file_ls: List[dict] = None,
    is_recursive: bool = True,
    parent_ls: Optional[List[str]] = None,
    just_folders: bool = False,
):  # recursive function that retrieves all files in subfolders

    if file_ls is None:
        file_ls = []

    parent_ls = parent_ls.copy() if parent_ls else [folder_id]

    new_files = await get_folder_contents_by_id(
        service, folder_id, parent_ls=parent_ls, just_folders=just_folders
    )
    file_ls.extend(new_files)

    if new_files and is_recursive:
        await ce.gather_with_concurrency(
            *[
                walk_folder_and_flatten(
                    service=service,
                    folder_id=file_obj["id"],
                    file_ls=file_ls,
                    parent_ls=[
                        *parent_ls,
                        *[
                            parent
                            for parent in file_obj["parent_ls"]
                            if parent not in parent_ls
                        ],
                    ],
                    just_folders=just_folders,
                )
                for file_obj in new_files
                if file_obj.get("mimeType") == "application/vnd.google-apps.folder"
            ],
            n=5
        )

    return file_ls

# %% ../../nbs/google/GDrive.ipynb 11
async def walk_folder(
    service: ga.Resource,
    folder_id: str,
    folder_obj: dict = None,
    parent_ls: List[str] = None,
    is_recursive: bool = True,
):  # recursive function that retrieves all files in subfolders

    parent_ls = parent_ls.copy() if parent_ls else [folder_id]

    folder_obj = folder_obj or (
        await get_file_by_id(service=service, file_id=folder_id, parent_ls=parent_ls)
    )

    folder_obj["children"] = (
        await get_folder_contents_by_id(
            service, folder_id, just_folders=False, parent_ls=parent_ls
        )
    ) or []

    if not is_recursive:
        return folder_obj

    for child in folder_obj["children"]:
        if child and child.get("mimeType") == "application/vnd.google-apps.folder":
            await walk_folder(
                service=service,
                folder_obj=child,
                folder_id=child.get("id"),
                parent_ls=parent_ls,
                is_recursive=is_recursive,
            )

    return folder_obj

# %% ../../nbs/google/GDrive.ipynb 13
class GDrive_MimeType_Enum(Enum):
    """for translating Google Drive types into 'exports as'"""

    docx = "application/vnd.openxmlformats-officedocument.wordprocessingml.document"
    pptx = "application/vnd.openxmlformats-officedocument.presentationml.presentation"

    odt = "application/vnd.oasis.opendocument.text"
    rtf = "application/rtf"
    pdf = "application/pdf"
    txt = "text/plain"

    zip_file = "application/zip"
    epub = "application/epub+zip"

    folder = "application/vnd.google-apps.folder"
    gdoc = "application/vnd.google-apps.document"
    gsheet = "application/vnd.google-apps.spreadsheet"
    gshortcut = "application/vnd.google-apps.shortcut"
    gslides = "application/vnd.google-apps.presentation"

    markdown = "text/markdown"

    # html = 'application/zip'
    # 'tex'  = 'application/zip'
    # 'html.zip'= 'application/zip'

# %% ../../nbs/google/GDrive.ipynb 14
class GDrive_ExportError(Exception):
    def __init__(self, doc_url):
        message = f"failure to download content for {doc_url}"
        super().__init__(message)


async def export_file_as(
    service: ga.Resource,
    file_id: str,
    export_mime_type: GDrive_MimeType_Enum,
    replace_folder: bool = True,
    export_fn: Callable = None,
    output_path: str = None,
):

    retry = 0
    content = None
    while retry < 3 and content is None:
        retry += 1
        try:
            content = (
                service.files()
                .export(fileId=file_id, mimeType=export_mime_type.value)
                .execute()
            )

        except HttpError as err:
            print(err)

        except Exception as e:
            print(e)

    if not content:
        raise GDrive_ExportError(file_id)

    if not export_fn:
        return content

    upsert_folder(output_path, replace_folder=replace_folder)

    return export_fn(content=content, output_folder=output_path)


async def export_as_pptx(service: ga.Resource, file_id: str, output_path: str):
    return await export_file_as(
        service=service,
        file_id=file_id,
        export_mime_type=GDrive_MimeType_Enum.pptx,
        export_fn=download_pptx,
        output_path=output_path,
    )


async def export_as_zip(service: ga.Resource, file_id: str, output_path: str):
    return await export_file_as(
        service=service,
        file_id=file_id,
        export_mime_type=GDrive_MimeType_Enum.zip_file,
        export_fn=download_zip,
        output_path=output_path,
    )


class ExportAs_Enum(Enum):
    pptx = partial(export_as_pptx)
    zip_file = partial(export_as_zip)

    @classmethod
    def _missing_(cls, value):
        value = value.lower()
        try:
            for member in cls:
                if member.name == value:
                    return member

        except:
            return cls.zip_file

        return cls.zip_file

# %% ../../nbs/google/GDrive.ipynb 18
class GDrive_ServiceRequired(Exception):
    def __init__(self, doc_url=None, doc_id=None):
        message = f"service (googleclientapi.discovery.Resource) required to download {doc_url or doc_id}"
        super().__init__(message)


@dataclass
class GDrive_File:
    id: str
    name: str

    modified_time: dt.datetime
    url: str

    auth: ga.GoogleAuth = field(repr=False)

    mime_str: str

    service: Any = field(repr=False, default=None)

    gdrive_mime_type: GDrive_MimeType_Enum = None

    parent_ls: List[str] = field(default=None)

    # gdrive_parents: List[Any] = field(default=None, repr=False)

    file: Any = field(repr=False, default=None)

    def _parent_init(self):
        if not self.service:
            self.service = auth.generate_gdrive_service()

        try:
            self.gdrive_mime_type = GDrive_MimeType_Enum(self.mime_str)

        except Exception as e:
            print(e)

    def __post_init__(self):
        self._parent_init()

    @classmethod
    async def _from_json(
        cls,
        gdrive_obj: dict,
        auth: ga.GoogleAuth,
        service: ga.Resource = None,
        **kwargs,
    ):
        gfile = cls(
            auth=auth,
            service=service or auth.generate_gdrive_service(),
            file=gdrive_obj,
            mime_str=gdrive_obj["mimeType"],
            url=str(gdrive_obj["webViewLink"]),
            id=gdrive_obj["id"],
            name=gdrive_obj["name"],
            parent_ls=gdrive_obj["parent_ls"],
            modified_time=convert_str_to_date(gdrive_obj["modifiedTime"]),
            **kwargs,
        )

        # if gfile.parent_ls:
        #     gfile.gdrive_parents = await ce.gather_with_concurrency(
        #         *[
        #             GDrive_File.get_by_id(
        #                 id=parent, auth=gfile.auth, service=gfile.service
        #             )
        #             for parent in gfile.parent_ls
        #         ],
        #         n=5,
        #     )

        return gfile

    @classmethod
    async def _super_get_by_id(
        cls,
        id: str,
        fields: List[str] = None,
        auth: ga.GoogleAuth = None,
        service: ga.Resource = None,
        return_raw: bool = False,
    ):
        fields = fields or DEFAULT_GDRIVE_FIELDS

        service = service or auth.generate_gdrive_service()

        file = await get_file_by_id(file_id=id, service=service, fields=fields)

        if return_raw:
            return file

        return await cls._from_json(gdrive_obj=file, auth=auth, service=service)

    @classmethod
    async def get_by_id(
        cls,
        id: str,
        auth: ga.GoogleAuth = None,
        service: ga.Resource = None,
        return_raw: bool = False,
        fields: List[str] = None,
    ):

        return await cls._super_get_by_id(
            id=id, auth=auth, service=service, return_raw=return_raw, fields=fields
        )

    async def export(self, output_path: str, export_mime_type: GDrive_MimeType_Enum):

        if isinstance(export_mime_type, GDrive_MimeType_Enum):
            export_mime_type = export_mime_type.name

            return await ExportAs_Enum(export_mime_type).value(
                service=self.service, file_id=self.id, output_path=output_path
            )

        return await ExportAs_Enum(export_mime_type).value(
            service=self.service, file_id=self.id, output_path=output_path
        )

# %% ../../nbs/google/GDrive.ipynb 23
@dataclass
class GDrive_Folder(GDrive_File):

    gdrive_files: List[GDrive_File] = field(default=None)

    async def _get_child_from_json(self, child, is_recursive: bool = True):
        if not child.get("is_folder"):
            g_child = await GDrive_File._from_json(
                gdrive_obj=child, service=self.service, auth=self.auth
            )

        if child.get("is_folder"):
            g_child = await GDrive_Folder._from_json(
                gdrive_obj=child, service=self.service, auth=self.auth
            )

            if is_recursive:
                await g_child.get_files(is_recursive=is_recursive)

        if not self.gdrive_files:
            self.gdrive_files = []

        self.gdrive_files.append(g_child)

    async def get_files(self, is_recursive: bool = True, return_raw: bool = False):

        folder_obj = await walk_folder(
            service=self.service,
            folder_id=self.id,
            is_recursive=is_recursive,
        )

        if return_raw:
            return folder_obj

        await ce.gather_with_concurrency(
            *[
                self._get_child_from_json(child=child, is_recursive=is_recursive)
                for child in folder_obj.get("children")
            ],
            n=10
        )

        return self.gdrive_files



"""Fill in a module description here"""

# AUTOGENERATED! DO NOT EDIT! File to edit: ../../nbs/google/GAuth.ipynb.

# %% auto 0
__all__ = ['DEFAULT_SCOPE', 'GoogleAuth_NoCreds_Error', 'generate_creds_from_file', 'GAuth_Required', 'GoogleAuth_NoEnv_Error',
           'generate_creds_from_env', 'GoogleAuth']

# %% ../../nbs/google/GAuth.ipynb 3
from google.auth.transport.requests import Request

# %% ../../nbs/google/GAuth.ipynb 4
import os

from dataclasses import dataclass, field
from typing import List

import dotenv


from googleapiclient.discovery import build
from google_auth_oauthlib.flow import InstalledAppFlow

import domolibrary_extensions.utils.convert as dtcv
import domolibrary_extensions.utils.files as dtfi

# %% ../../nbs/google/GAuth.ipynb 5
from google.oauth2.credentials import Credentials
from googleapiclient.discovery import Resource

# %% ../../nbs/google/GAuth.ipynb 6
class GoogleAuth_NoCreds_Error(Exception):
    def __init__(self, message: str = None):
        super().__init__(
            message
            or "unable to generate a creds obj, must pass token_path or means of generating creds"
        )


DEFAULT_SCOPE = [
    "https://www.googleapis.com/auth/drive",
    "https://www.googleapis.com/auth/drive.file",
    "https://www.googleapis.com/auth/drive.metadata.readonly",
    "https://www.googleapis.com/auth/drive.readonly",
]

# %% ../../nbs/google/GAuth.ipynb 7
def _test_or_generate_new_creds(creds, flow, port=0):
    if not creds and not flow:
        raise GoogleAuth_NoCreds_Error()

    if creds and creds.valid:
        return creds

    if creds and creds.expired and creds.refresh_token:
        creds.refresh(Request())
        return creds

    return flow.run_local_server(port=port)

# %% ../../nbs/google/GAuth.ipynb 9
def generate_creds_from_file(
    client_path: str,
    token_path: str = "./env_token.json",
    scope: List[str] = None,
    port=0,
) -> Credentials:

    scope = scope or DEFAULT_SCOPE

    creds = None

    if token_path and os.path.exists(token_path):
        creds = Credentials.from_authorized_user_file(token_path, scope)

    if not creds and (not client_path or not os.path.exists(client_path)):
        raise FileExistsError(client_path)

    flow = InstalledAppFlow.from_client_secrets_file(client_path, scope)

    creds = _test_or_generate_new_creds(creds=creds, flow=flow, port=port)

    dtfi.upsert_file(
        file_path=token_path,
        content=creds.to_json(),
    )

    return creds

# %% ../../nbs/google/GAuth.ipynb 11
class GAuth_Required(Exception):
    def __init__(self):
        message = f"must pass valid Google Auth or Service with `service_name = docs`"
        super().__init__(message)


class GoogleAuth_NoEnv_Error(Exception):
    def __init__(self, env_token):
        super().__init__(f"{env_token} not found in ENV")


def generate_creds_from_env(
    client_env_key, env_path, token_env_key=None, scope: List[str] = None, port=0
):

    scope = scope or DEFAULT_SCOPE

    if env_path:
        assert os.path.exists(env_path)
        print(f"reloading env - {env_path}")
        dotenv.load_dotenv(env_path, override=True)

    token_obj = None
    creds = None

    if token_env_key:
        token_obj = dtcv.convert_json_to_obj(os.getenv(token_env_key))

        if token_obj:
            creds = Credentials.from_authorized_user_info(token_obj, scope)

    if not creds and not os.getenv(client_env_key):
        raise GoogleAuth_NoEnv_Error(client_env_key)

    client_obj = dtcv.convert_json_to_obj(os.getenv(client_env_key))

    flow = InstalledAppFlow.from_client_config(client_obj, scope)

    creds = _test_or_generate_new_creds(
        creds=creds,
        flow=flow,
        port=port,
    )

    dtfi.update_env(env_key=token_env_key, content=creds.to_json(), env_path=env_path)

    return creds

# %% ../../nbs/google/GAuth.ipynb 14
@dataclass
class GoogleAuth:
    creds: Credentials

    scope: List[str] = field(default_factory=lambda: DEFAULT_SCOPE)

    @classmethod
    def generate_from_env(
        cls,
        env_path,
        client_env_key,
        token_env_key=None,
        scope: List[str] = None,
    ):
        """
        credentials should be stored in the .env file as a json (NOT A STRING)
        token will be stored as a string (NOT A JSON OBJ)

        generate creds here, https://console.cloud.google.com/apis/credentials/oauthclient
        for a new implementation you'll need to
        1. set up a project at console.cloud.google.com,
        2. enable (turn on) the appropriate APIs for that project
        3. configure OAuth access (your python script will use an Oauth token to authenticate, so your project needs the OAuth screen)
        4. download the creds file and store it as env_creds.json
        """

        creds = generate_creds_from_env(
            client_env_key=client_env_key,
            token_env_key=token_env_key,
            scope=scope,
            env_path=env_path,
        )

        return cls(creds=creds)

    @classmethod
    def generate_from_file(
        cls, client_path: str, token_path: str, scope: List[str] = None
    ):
        """generate creds here, https://console.cloud.google.com/apis/credentials/oauthclient
        for a new implementation you'll need to
        1. set up a project at console.cloud.google.com,
        2. enable (turn on) the appropriate APIs for that project
        3. configure OAuth access (your python script will use an Oauth token to authenticate, so your project needs the OAuth screen)
        4. download the client json file and store it as env_creds.json
        """

        creds = generate_creds_from_file(
            client_path=client_path, token_path=token_path, scope=scope
        )

        return cls(creds=creds)

    def _generate_service(self, service, version) -> Resource:
        """Generates a service object for the given service name and version"""

        print(
            f"generating {service} - {version} service object on {self.__class__.__name__}"
        )

        return build(serviceName=service, version=version, credentials=self.creds)

    def generate_gdrive_service(self) -> Resource:
        """Generates a service object for the given service name and version"""

        return self._generate_service(service="drive", version="v3")

    def generate_gdocs_service(self):
        return self._generate_service(service="docs", version="v1")

    def generate_sheets_service(self):
        return self._generate_service(service="sheets", version="v4")



# AUTOGENERATED! DO NOT EDIT! File to edit: ../../nbs/scraper/ScrapeDomoSupport.ipynb.

# %% ../../nbs/scraper/ScrapeDomoSupport.ipynb 2
from __future__ import annotations
from functools import partial

from selenium.webdriver.common.keys import Keys

import json
from io import StringIO
import datetime as dt
import time

import pandas as pd
from bs4 import BeautifulSoup

import domolibrary_extensions.scraper.routes as desr
import domolibrary_extensions.scraper.Scraper as desc
import domolibrary_extensions.utils.web as deuw

# %% auto 0
__all__ = ['domo_support_link_extractor_fn', 'DomoSupport_ScrapeConfig_SupportAllTickets', 'DomoSupport_ScrapeConfig_Ticket',
           'DomoSupport_ScrapeConfig_TicketComments', 'DomoSupport_ScrapeFactory',
           'domo_support_all_tickets_pagesource', 'domo_support_all_tickets_extract_content',
           'domo_support_link_extractor_fn__custom', 'domo_support_generate_filename',
           'domo_support_ticket_extract_content', 'domo_support_ticket_comments_extract_content']

# %% ../../nbs/scraper/ScrapeDomoSupport.ipynb 3
from domolibrary_extensions.utils.web import (
    extract_links,
    generate_filename_from_url,
    remove_query_params_from_url,
)

# %% ../../nbs/scraper/ScrapeDomoSupport.ipynb 7
def domo_support_all_tickets_pagesource(
    driver: desr.webdriver,  # must pass an authenticated driver
    return_entire_page: bool = True,
    return_soup: bool = False,
    **kwargs,
) -> BeautifulSoup:
    url = f"https://domo-support.domo.com/s/"

    driver.get(url)

    time.sleep(2)

    # browser element
    page_source = desr.get_pagesource(
        driver=driver,
        search_criteria_tuple=(desr.By.CLASS_NAME, "cellContainer"),
        return_soup=False,
        return_entire_page=True,
    )

    new_len_table = 0
    prev_len_table = None

    while new_len_table != prev_len_table:
        scroll_bar = desr.wait_and_return(
            driver=driver,
            search_criteria_tuple=(desr.By.CLASS_NAME, "uiScroller"),
            return_soup=False,
            # min_sleep_time=15,
        )
        scroll_bar.send_keys(Keys.END)

        time.sleep(2)
        print(f"💤📜 sleeping on the scroll {new_len_table} 💤📜")

        soup = BeautifulSoup(page_source, parser="lxml")

        prev_len_table = new_len_table
        new_len_table = len(soup.find_all("td"))

    print(f"🚀 done retrieving table {new_len_table} 🚀")

    return desr.get_pagesource(
        driver=driver,
        search_criteria_tuple=(desr.By.TAG_NAME, "table"),
        return_soup=return_soup,
        return_entire_page=return_entire_page,
    )

# %% ../../nbs/scraper/ScrapeDomoSupport.ipynb 10
def domo_support_all_tickets_extract_content(
    soup,
    return_soup: bool = False,
    return_extractor: bool = False,
    base_url="https://domo-support.domo.com",
) -> pd.DataFrame:
    soup = soup.find("table")

    if return_soup:
        return soup

    df_with_links = pd.read_html(StringIO(str(soup)), extract_links="body")[0]

    df = pd.read_html(StringIO(str(soup)))[0]

    rename_cols = {
        col: col.replace("Column Actions", "")
        .replace("Sorted", "")
        .replace("Sort", "")
        .replace("Ascending", "")
        .replace("Descending", "")
        .strip()
        for col in df.columns
    }

    df.rename(columns=rename_cols, inplace=True)
    df_with_links.rename(columns=rename_cols, inplace=True)

    df_with_links = pd.DataFrame(
        df_with_links["Case Number"].tolist(),
        index=df_with_links.index,
        columns=["Case Number", "URL"],
    )

    df["Case Number"] = df["Case Number"].astype(int)
    df_with_links["Case Number"] = df_with_links["Case Number"].astype(int)

    df = pd.merge(df, df_with_links, on="Case Number", how="inner")

    df["URL"] = base_url + df["URL"]

    print(f"🚀 {len(df.index)} tickets retrieved")

    if return_extractor:
        return df.to_csv(), "content.csv"

    return df

# %% ../../nbs/scraper/ScrapeDomoSupport.ipynb 12
def domo_support_link_extractor_fn__custom(url, base_url, debug_prn: bool = False):
    url = deuw.remove_query_params_from_url(url)

    if debug_prn:
        print(url)

    if not url or "/s/case/" not in url:
        return None

    url = "/".join(url.split("/")[:6])

    if url.endswith("/"):
        url = url[:-1]

    return url


domo_support_link_extractor_fn = partial(
    extract_links,
    custom_link_extractor_fn=partial(
        domo_support_link_extractor_fn__custom, debug_prn=False
    ),
)


def domo_support_generate_filename(
    domo_instance,
    variant=None,
    url=None,
    file_name="index.html",
    download_folder="./SCRAPE",
    **kwargs,
):
    """must apply at scrape_crawler definition"""

    base = f"{download_folder}/{domo_instance}"
    if variant:
        return f"{base}/{variant}/{str(dt.date.today())}/{file_name}"

    return f"{base}/{generate_filename_from_url(url)}/{file_name}"

# %% ../../nbs/scraper/ScrapeDomoSupport.ipynb 14
DomoSupport_ScrapeConfig_SupportAllTickets = desc.ScrapeConfig(
    id=r".*/domo-support.domo.com/s/$",
    link_extractor_fn=domo_support_link_extractor_fn,
    content_extractor_fn=partial(
        domo_support_all_tickets_extract_content, return_extractor=True
    ),
    pagesource_fn=partial(
        domo_support_all_tickets_pagesource, return_entire_page=True, return_soup=False
    ),
    generate_filename_fn=partial(
        domo_support_generate_filename, variant="all_tickets", domo_instance="default"
    ),
)

# %% ../../nbs/scraper/ScrapeDomoSupport.ipynb 17
def domo_support_ticket_extract_content(
    soup,
    return_soup: bool = False,
    return_extractor: bool = False,
    base_url="https://domo-support.domo.com",
) -> pd.DataFrame:

    table_soup = soup.find("div", class_="test-id__record-layout-container")

    if return_soup:
        return table_soup

    obj = {
        row.find("div", class_="slds-form-element__label")
        .text.strip(): row.find("div", class_="slds-form-element__control")
        .text.strip()
        for row in table_soup.find_all("div", class_="slds-form-element")
    }

    if return_extractor:
        return (json.dumps(obj), "ticket_meta.json")

    return obj


DomoSupport_ScrapeConfig_Ticket = desc.ScrapeConfig(
    id=r".*/s/case/*",
    search_criteria_tuple=(desr.By.CLASS_NAME, "slds-form-element"),
    link_extractor_fn=domo_support_link_extractor_fn,
    content_extractor_fn=partial(
        domo_support_ticket_extract_content, return_extractor=True
    ),
    generate_filename_fn=partial(
        domo_support_generate_filename, domo_instance="default"
    ),
)

# %% ../../nbs/scraper/ScrapeDomoSupport.ipynb 20
def domo_support_ticket_comments_extract_content(
    soup,
    return_soup: bool = False,
    return_extractor: bool = False,
    base_url="https://domo-support.domo.com",
):

    # description_class = "test-id__field-value"
    soup = soup.find(class_="listContent")

    if return_soup:
        return soup

    casecomments_ls = []
    for li in soup.find("ul").find_all("li"):
        comment_hdr = li.find(class_="date").text
        comment_body = li.find(class_="commentBody").text

        casecomments_ls.append({"date": comment_hdr, "body": comment_body})

    if return_extractor:
        res = {"case_comments": casecomments_ls}
        return (json.dumps(res), "ticket_comments.json")

    return pd.DataFrame(casecomments_ls)


DomoSupport_ScrapeConfig_TicketComments = desc.ScrapeConfig(
    id=r".*/s/case/*",
    link_extractor_fn=domo_support_link_extractor_fn,
    search_criteria_tuple=(desr.By.CLASS_NAME, "commentBody"),
    content_extractor_fn=partial(
        domo_support_ticket_comments_extract_content, return_extractor=True
    ),
    generate_filename_fn=partial(
        domo_support_generate_filename, domo_instance="default"
    ),
)

# %% ../../nbs/scraper/ScrapeDomoSupport.ipynb 23
DomoSupport_ScrapeFactory = desc.ScrapeFactory(
    factory_configs=[
        DomoSupport_ScrapeConfig_SupportAllTickets,
        DomoSupport_ScrapeConfig_Ticket,
        DomoSupport_ScrapeConfig_TicketComments,
    ]
)



"""supporting functions"""

# AUTOGENERATED! DO NOT EDIT! File to edit: ../../nbs/utils/execution.ipynb.

# %% auto 0
__all__ = ['auto_retry_sync', 'auto_retry_async', 'gather_with_concurrency']

# %% ../../nbs/utils/execution.ipynb 3
from typing import Callable, Any, List

from functools import wraps

import asyncio
import time

# %% ../../nbs/utils/execution.ipynb 5
def auto_retry_sync(
    max_retry=1, exceptions_to_ignore: List[Exception] = None, sleep_beween_retry=5
):

    exceptions_to_ignore = exceptions_to_ignore or []

    def actual_decorator(func):
        @wraps(func)
        def wrapper(*args, **kwargs):

            retry = 0

            while retry <= max_retry:
                try:
                    return func(*args, **kwargs)

                except Exception as e:
                    if any((isinstance(e, ig_ex) for ig_ex in exceptions_to_ignore)):
                        raise e

                    if retry >= max_retry:
                        raise e

                    retry += 1

                    print(f"retry decorator attempt - {retry}, {e}")
                    time.sleep(sleep_beween_retry)

        return wrapper

    return actual_decorator


def auto_retry_async(
    max_retry=1, exceptions_to_ignore: List[Exception] = None, sleep_beween_retry=5
):
    exceptions_to_ignore = exceptions_to_ignore or []

    def actual_decorator(func):
        @wraps(func)
        async def wrapper(*args, **kwargs):

            retry = 0

            while retry <= max_retry:
                try:
                    return await func(*args, **kwargs)

                except Exception as e:
                    if any((isinstance(e, ig_ex) for ig_ex in exceptions_to_ignore)):
                        raise e

                    if retry >= max_retry:
                        raise e

                    retry += 1

                    print(f"retry decorator attempt - {retry}, {e}")
                    await asyncio.sleep(sleep_beween_retry)

        return wrapper

    return actual_decorator

# %% ../../nbs/utils/execution.ipynb 10
async def gather_with_concurrency(
    *coros,  # list of coroutines to await
    n=60,  # number of open coroutines
):
    """limits the number of open coroutines at a time."""

    semaphore = asyncio.Semaphore(n)

    async def sem_coro(coro):
        async with semaphore:
            return await coro

    return await asyncio.gather(*(sem_coro(c) for c in coros))



# AUTOGENERATED! DO NOT EDIT! File to edit: ../../nbs/jira/routes.ipynb.

# %% auto 0
__all__ = ['get_boards', 'get_board_epics', 'get_epic', 'search_issues_jql', 'get_issue_by_id', 'get_board_issues',
           'get_issue_remotelinks', 'upsert_issue_remotelink', 'get_issue_links', 'get_issue_comments']

# %% ../../nbs/jira/routes.ipynb 2
import httpx

import urllib

import domolibrary_extensions.client as gd
import domolibrary_extensions.jira.auth as ja

# %% ../../nbs/jira/routes.ipynb 5
async def get_boards(
    auth: ja.JiraAuth,
    params: dict = None,
    debug_api: bool = False,
    debug_loop: bool = False,
    return_raw: bool = False,
    client: httpx.AsyncClient = None,
    is_ignore_cache: bool = False,
    cache_path: str = None,
    timeout: int = 30,
):
    url = f"https://{auth.instance}/rest/agile/1.0/board"

    def arr_fn(res):
        return res.response["values"]

    client = client or httpx.AsyncClient()

    res = await gd.looper(
        url=url,
        auth=auth,
        arr_fn=arr_fn,
        params=params,
        method="GET",
        debug_loop=debug_loop,
        debug_api=debug_api,
        client=client,
        is_ignore_cache=is_ignore_cache,
        cache_path=cache_path,
        timeout=timeout,
        return_raw=return_raw,
    )

    return res

# %% ../../nbs/jira/routes.ipynb 11
async def get_board_epics(
    auth: ja.JiraAuth,
    board_id: int,
    params: dict = None,
    is_ignore_cache: bool = False,
    cache_path: str = None,
    client: httpx.AsyncClient = None,
    timeout: int = 30,
    debug_api: bool = False,
    debug_loop: bool = False,
    return_raw: bool = False,
):
    url = f"https://{auth.instance}/rest/agile/1.0/board/{board_id}/epic"

    def arr_fn(res):
        return res.response["values"]

    res = await gd.looper(
        method="get",
        url=url,
        auth=auth,
        params=params,
        debug_api=debug_api,
        debug_loop=debug_loop,
        arr_fn=arr_fn,
        is_ignore_cache=is_ignore_cache,
        cache_path=cache_path,
        client=client,
        timeout=timeout,
        return_raw=return_raw,
    )

    if not res.is_success:
        raise gd.BaseError(res=res)

    return res

# %% ../../nbs/jira/routes.ipynb 14
async def get_epic(
    epic_id,
    auth: ja.JiraAuth,
    debug_api: bool = False,
    is_ignore_cache: bool = False,
    cache_path: str = None,
    client: httpx.AsyncClient = None,
    timeout: int = 30,
):
    url = f"https://{auth.instance}/rest/agile/1.0/epic/{epic_id}"

    res = await gd.get_data(
        method="get",
        url=url,
        auth=auth,
        debug_api=debug_api,
        is_ignore_cache=is_ignore_cache,
        cache_path=cache_path,
        client=client,
        timeout=timeout,
    )

    if not res.is_success:
        raise gd.BaseError(res=res)

    return res

# %% ../../nbs/jira/routes.ipynb 18
async def search_issues_jql(
    auth: ja.JiraAuth,
    jql: str,
    params: dict = None,
    debug_api: bool = False,
    return_raw: bool = False,
    client: httpx.AsyncClient = None,
    is_ignore_cache: bool = False,
    cache_path: str = None,
):
    params = params or {}

    url = f"https://{auth.instance}/rest/api/2/search?"

    res = await gd.get_data(
        url=url,
        auth=auth,
        method="GET",
        debug_api=debug_api,
        params={**params, "jql": urllib.parse.quote(jql)},
        client=client,
        is_ignore_cache=is_ignore_cache,
        cache_path=cache_path,
    )

    if not res.is_success:
        raise gd.BaseError(res=res)

    if return_raw:
        return res

    res.response = res.response["issues"]

    return res

# %% ../../nbs/jira/routes.ipynb 19
async def get_issue_by_id(
    auth: ja.JiraAuth,
    issue_id: int,
    debug_api: bool = False,
    is_ignore_cache: bool = False,
    cache_path: str = None,
    client: httpx.AsyncClient = None,
    timeout: int = 30,
):
    url = f"https://{auth.instance}/rest/api/2/issue/{issue_id}"

    res = await gd.get_data(
        method="get",
        url=url,
        auth=auth,
        debug_api=debug_api,
        is_ignore_cache=is_ignore_cache,
        cache_path=cache_path,
        client=client,
        timeout=timeout,
    )

    if not res.is_success:
        raise gd.BaseError(res=res)

    return res

# %% ../../nbs/jira/routes.ipynb 20
async def get_board_issues(
    auth: ja.JiraAuth,
    board_id: int,
    params: dict = None,
    debug_api: bool = False,
    debug_loop: bool = False,
    return_raw: bool = False,
    is_ignore_cache: bool = False,
    cache_path: str = None,
    client: httpx.AsyncClient = None,
    timeout: int = 30,
):
    url = f"https://{auth.instance}/rest/agile/1.0/board/{board_id}/issue"

    def arr_fn(res):
        return res.response["issues"]

    res = await gd.looper(
        method="get",
        url=url,
        auth=auth,
        params=params,
        offset_params={"offset": "startAt", "limit": "maxResults"},
        offset_params_is_header=True,
        debug_api=debug_api,
        debug_loop=debug_loop,
        arr_fn=arr_fn,
        is_ignore_cache=is_ignore_cache,
        cache_path=cache_path,
        client=client,
        timeout=timeout,
        return_raw=return_raw,
    )

    if not res.is_success:
        raise gd.BaseError(res=res)

    return res

# %% ../../nbs/jira/routes.ipynb 21
async def get_issue_remotelinks(
    auth: ja.JiraAuth,
    issue_id: int,
    debug_api: bool = False,
    is_ignore_cache: bool = False,
    cache_path: str = None,
    client: httpx.AsyncClient = None,
    timeout: int = 30,
):
    # url = f'https://{auth.instance}/rest/api/2/issueLink'
    url = f"https://{auth.instance}/rest/api/2/issue/{issue_id}/remotelink"

    res = await gd.get_data(
        method="get",
        url=url,
        auth=auth,
        debug_api=debug_api,
        is_ignore_cache=is_ignore_cache,
        cache_path=cache_path,
        client=client,
        timeout=timeout,
    )

    return res

# %% ../../nbs/jira/routes.ipynb 22
async def upsert_issue_remotelink(
    issue_id,
    body,
    auth: ja.JiraAuth,
    debug_api: bool = False,
    is_ignore_cache: bool = False,
    cache_path: str = None,
    client: httpx.AsyncClient = None,
    timeout: int = 30,
):

    url = f"https://{auth.instance}/rest/api/latest/issue/{issue_id}/remotelink"

    res = await gd.get_data(
        method="post",
        url=url,
        body=body,
        auth=auth,
        headers={"Content-Type": "application/json"},
        debug_api=debug_api,
        is_ignore_cache=is_ignore_cache,
        cache_path=cache_path,
        client=client,
        timeout=timeout,
    )

    return res

# %% ../../nbs/jira/routes.ipynb 23
async def get_issue_links(
    auth,
    issue_id: int,
    debug_api: bool = False,
    return_raw: bool = False,
    is_ignore_cache: bool = False,
    cache_path: str = None,
    client: httpx.AsyncClient = None,
    timeout: int = 30,
):
    url = f"https://{auth.instance}/rest/api/2/issue/{issue_id}?fields=issuelinks,customfield_10940"
    # ,parentIssuesOf

    res = await gd.get_data(
        method="get",
        url=url,
        auth=auth,
        debug_api=debug_api,
        is_ignore_cache=is_ignore_cache,
        cache_path=cache_path,
        client=client,
        # body = {"expand" : 'links'},
        timeout=timeout,
    )

    if return_raw:
        return res

    fields = res.response["fields"]

    parent_id = fields.get("customfield_10940")

    issuelinks = fields["issuelinks"]

    il = [
        {
            "url": f"https://{auth.instance}/browse/{link['inwardIssue']['key']}",
            "id": link["inwardIssue"]["key"],
            "type": link["type"]["inward"],
            "status": link["inwardIssue"]["fields"]["status"]["name"],
            "title": link["inwardIssue"]["fields"]["summary"],
        }
        for link in issuelinks
    ]

    res.response = {
        "parent_epic": (
            {"id": parent_id, "url": f"https://{auth.instance}/browse/{parent_id}"}
            if parent_id
            else None
        ),
        "issue_links": il,
    }

    return res

# %% ../../nbs/jira/routes.ipynb 24
async def get_issue_comments(
    issue_id,
    auth: ja.JiraAuth,
    debug_api: bool = False,
    return_raw: bool = False,
    is_ignore_cache: bool = False,
    cache_path: str = None,
    client: httpx.AsyncClient = None,
    timeout: int = 30,
):
    url = (
        f"https://{auth.instance}/rest/api/2/issue/{issue_id}/comment?orderBy=-created&maxResults=0&expand=renderedBody"
        if auth.is_datacenter
        else f"https://{auth.instance}/rest/api/2/issue/{issue_id}/comment?orderBy=-created&maxResults=0&expand=renderedBody"
    )

    def arr_fn(res):
        return res.response["comments"]

    res = await gd.looper(
        method="get",
        url=url,
        auth=auth,
        arr_fn=arr_fn,
        offset_params={"offset": "startAt", "limit": "maxResults"},
        offset_params_is_header=True,
        debug_api=debug_api,
        is_ignore_cache=is_ignore_cache,
        cache_path=cache_path,
        client=client,
        timeout=timeout,
        return_raw=return_raw,
    )

    return res



# AUTOGENERATED! DO NOT EDIT! File to edit: ../../nbs/utils/xkcd_pass.ipynb.

# %% auto 0
__all__ = ['LEET', 'PADDING', 'add_leet_to_string', 'add_padding_characters_fn', 'process_add_leet', 'process_pad_suffix_fn',
           'process_random_capitalization_fn', 'process_first_capitalization_fn',
           'process_caps_first_word_add_year_and_add_suffix', 'generate_xkcd_password', 'generate_domo_password',
           'process_domo_password_fn']

# %% ../../nbs/utils/xkcd_pass.ipynb 3
from xkcdpass import xkcd_password as xp
import random
from typing import Callable
import datetime as dt
import nbdev

# %% ../../nbs/utils/xkcd_pass.ipynb 4
LEET = {"a": "@", "i": "!", "A": "@", "I": "!", "e": "3", "E": "3"}


def add_leet_to_string(text, leet):
    return "".join(leet.get(k, k) for k in text)

# %% ../../nbs/utils/xkcd_pass.ipynb 6
PADDING = ".!?"


def add_padding_characters_fn(
    text,
    padding,
    n: int = 1,  # num padding characters to add
):

    text += "".join(random.choices(padding, k=n))
    return text

# %% ../../nbs/utils/xkcd_pass.ipynb 9
def process_add_leet(my_pass, **kwarags):
    leet = LEET
    keep_loop = False

    if not any((l for l in my_pass if l in leet.keys())):
        keep_loop = True
        return my_pass, keep_loop

    my_pass = add_leet_to_string(my_pass, leet=leet)
    return my_pass, keep_loop


def process_pad_suffix_fn(my_pass):
    return add_padding_characters_fn(my_pass, padding=PADDING, n=1), False


def process_random_capitalization_fn(text, delimiter, **kwargs):
    if delimiter not in text:
        return text, False

    word_ls = text.split(delimiter)

    word_ls = [random.choice((str.upper, str.lower))(word) for word in word_ls]

    return delimiter.join(word_ls), False


def process_first_capitalization_fn(text, delimiter, **kwargs):
    if delimiter not in text:
        return text, False

    word_ls = text.split(delimiter)

    word_ls = [
        word.upper() if idx == 0 else word.lower() for idx, word in enumerate(word_ls)
    ]

    return delimiter.join(word_ls), False


def process_caps_first_word_add_year_and_add_suffix(my_pass, delimiter, **kwargs):
    word_ls = my_pass.split(delimiter)
    word_ls[0] = word_ls[0].upper()
    my_pass = delimiter.join(word_ls)

    my_pass += dt.date.today().strftime("%Y")
    my_pass, keep_loop = process_pad_suffix_fn(my_pass)

    return my_pass, keep_loop


def generate_xkcd_password(
    min_word_length=5, max_word_length=6, valid_chars="[a-z]", delimiter="-"
):
    wordfile = xp.locate_wordfile()
    mywords = xp.generate_wordlist(
        wordfile=wordfile,
        min_length=min_word_length,
        max_length=max_word_length,
        valid_chars=valid_chars,
    )
    return xp.generate_xkcdpassword(mywords, numwords=3, delimiter=delimiter)

# %% ../../nbs/utils/xkcd_pass.ipynb 11
def generate_domo_password(
    delimiter="-",
    process_fn: Callable = process_caps_first_word_add_year_and_add_suffix,
):
    keep_loop = True

    while keep_loop == True:
        keep_loop = False
        my_pass = generate_xkcd_password(
            min_word_length=5,
            max_word_length=6,
            valid_chars="[a-z]",
            delimiter=delimiter,
        )

        if process_fn:
            my_pass, keep_loop = process_fn(my_pass=my_pass, delimiter=delimiter)

        return my_pass

# %% ../../nbs/utils/xkcd_pass.ipynb 14
def process_domo_password_fn(my_pass, delimiter):
    my_pass, keep_loop = process_first_capitalization_fn(my_pass, delimiter)
    my_pass += dt.date.today().strftime("%Y")
    my_pass, keep_loop = process_pad_suffix_fn(my_pass)

    return my_pass, keep_loop


def generate_domo_password(delimiter="-", process_fn: Callable = None):

    wordfile = xp.locate_wordfile()
    mywords = xp.generate_wordlist(
        wordfile=wordfile, min_length=5, max_length=6, valid_chars="[a-z]"
    )

    my_pass = None
    keep_loop = True

    while keep_loop == True:
        my_pass = xp.generate_xkcdpassword(mywords, numwords=3, delimiter=delimiter)
        keep_loop = False

        if process_fn:
            my_pass, keep_loop = process_fn(my_pass=my_pass, delimiter=delimiter)

        return my_pass



# AUTOGENERATED! DO NOT EDIT! File to edit: ../../nbs/scraper/DriverGenerator.ipynb.

# %% auto 0
__all__ = ['DriverGenerator_NotInPath', 'DriverGenerator', 'DriverGeneratorLinux']

# %% ../../nbs/scraper/DriverGenerator.ipynb 2
import subprocess

from selenium import webdriver
from typing import Callable
import domolibrary_extensions.scraper.routes as desr

# %% ../../nbs/scraper/DriverGenerator.ipynb 5
class DriverGenerator_NotInPath(Exception):
    def __init__(self):
        super().__init__(
            "google-chrome not found in PATH and therefore cannot be accessed by a driver.  install it or add to PATH"
        )


class DriverGenerator:
    """
    crawler may request multiple drivers, so need a drivergenerator to spin up multiple
    class for generating drivers, can be extended for other drivers
    """

    driver_path: str = None
    authenticator_fn: Callable = (
        None  # if provided generated webdrivers will authenticate drivers
    )

    def __init__(self, driver_path=None, authenticator_fn: Callable = None):
        self.driver_path = driver_path
        self.authenticator_fn = authenticator_fn

    def get_webdriver(
        self,
        driver_path=None,
        is_headless: bool = True,
        authenticator_fn: Callable = None,
    ) -> webdriver:
        """create a chrome webdriver"""

        driver_path = driver_path or self.driver_path

        if not self.driver_path:
            self.driver_path = driver_path

        driver = desr.generate_webdriver(
            driver_path=driver_path, is_headless=is_headless
        )

        authenticator_fn = authenticator_fn or self.authenticator_fn
        if authenticator_fn:
            authenticator_fn(driver=driver)

        return driver

# %% ../../nbs/scraper/DriverGenerator.ipynb 7
class DriverGeneratorLinux(DriverGenerator):
    def __init__(self, driver_path, debug_prn: bool = False):
        self.driver_path = driver_path or "/usr//bin/chromedriver"

        assert self.is_chrome_in_path()

        if debug_prn:
            print(self.get_chromedriver_version())

        super().__init__()

    def is_chrome_in_path(self):
        result = subprocess.run(["which", "google-chrome"], stdout=subprocess.PIPE)
        is_chrome_in_path = result.stdout != b""

        if not is_chrome_in_path:
            raise DriverGenerator_NotInPath()

        return is_chrome_in_path

    def get_chromedriver_version(self):
        result = subprocess.run(["chromedriver", "--version"], stdout=subprocess.PIPE)
        return result.stdout.decode("utf-8")



# AUTOGENERATED! DO NOT EDIT! File to edit: ../../nbs/asana/01_user.ipynb.

# %% ../../nbs/asana/01_user.ipynb 2
from __future__ import annotations

from dataclasses import dataclass, field
from typing import List

from nbdev.showdoc import patch_to

import domolibrary_extensions.asana.auth as aa
import domolibrary_extensions.client as gd

# %% auto 0
__all__ = ['AsanaUser']

# %% ../../nbs/asana/01_user.ipynb 4
@dataclass
class AsanaUser:
    id: str
    name: str
    auth: aa.AsanaAuth = field(repr=False)
    resource_type: str = field(repr=False)
    email: str = None

    @classmethod
    def _from_json(cls, obj, auth: aa.AsanaAuth):
        return cls(
            id=obj["gid"],
            name=obj["name"],
            email=obj.get("email"),
            resource_type=obj["resource_type"],
            auth=auth,
        )

# %% ../../nbs/asana/01_user.ipynb 5
@patch_to(AsanaUser, cls_method=True)
async def get_users(
    cls, auth: aa.AsanaAuth, debug_api: bool = False, return_raw: bool = False
) -> List[AsanaUser]:
    url = f"{auth.base_url}/users/"

    res = await gd.get_data(url, auth=auth, method="GET", debug_api=debug_api)

    if not res.is_success:
        raise gd.BaseError(res=res)

    if return_raw:
        return res

    return [cls._from_json(user_obj, auth=auth) for user_obj in res.response["data"]]

# %% ../../nbs/asana/01_user.ipynb 7
@patch_to(AsanaUser, cls_method=True)
async def get_by_id(
    cls,
    user_id: str,
    auth: aa.AsanaAuth,
    debug_api: bool = False,
    return_raw: bool = False,
) -> AsanaUser:
    """Get a user by ID"""
    url = f"{auth.base_url}/users/{user_id}"

    res = await gd.get_data(url, auth=auth, method="GET", debug_api=debug_api)

    if not res.is_success:
        raise gd.BaseError(res=res)

    if return_raw:
        return res

    return cls._from_json(res.response["data"], auth=auth)



# AUTOGENERATED! DO NOT EDIT! File to edit: ../../nbs/asana/01_project.ipynb.

# %% ../../nbs/asana/01_project.ipynb 2
from __future__ import annotations
from typing import List, Any
from dataclasses import dataclass, field
from typing import List
import datetime as dt

from nbdev.showdoc import patch_to

import domolibrary_extensions.utils.utils as deut
import domolibrary_extensions.client as gd
import domolibrary_extensions.asana.auth as aa
import domolibrary_extensions.asana.user as au

# %% auto 0
__all__ = ['AsanaProject']

# %% ../../nbs/asana/01_project.ipynb 4
@dataclass
class AsanaProject:
    id: str
    name: str
    workspace_id: str
    permalink_url: str
    auth: aa.AsanaAuth = field(repr=False)

    is_archived: bool
    is_completed: bool

    created_date: dt.datetime
    modified_date: dt.datetime

    owner: au.AsanaUser
    members: List[au.AsanaUser]

    due_date: dt.datetime = None
    completed_date: dt.datetime = None

    asana_tasks: List[Any] = None  # AsanaTask

    @classmethod
    def _from_json(cls, obj, auth: aa.AsanaAuth):
        owner = (
            au.AsanaUser._from_json(obj.get("owner"), auth=auth)
            if obj.get("owner")
            else None
        )

        members = (
            [
                au.AsanaUser._from_json(member, auth=auth)
                for member in obj.get("members", [])
            ]
            if obj.get("members")
            else None
        )

        created_date = deut.convert_str_to_date(obj.get("created_at"))
        modified_date = deut.convert_str_to_date(obj.get("modified_at"))
        due_date = deut.convert_str_to_date(obj.get("due_on"))
        completed_date = deut.convert_str_to_date(obj.get("completed_at"))

        return cls(
            auth=auth,
            id=obj.get("gid"),
            name=obj.get("name"),
            workspace_id=obj.get("workspace", {}).get("gid"),
            permalink_url=obj.get("permalink_url"),
            is_archived=obj.get("archived"),
            is_completed=obj.get("completed"),
            created_date=created_date,
            modified_date=modified_date,
            owner=owner,
            members=members,
            due_date=due_date,
            completed_date=completed_date,
        )

# %% ../../nbs/asana/01_project.ipynb 5
@patch_to(AsanaProject, cls_method=True)
async def get_projects(
    cls: AsanaProject,
    auth: aa.AsanaAuth,
    debug_api: bool = False,
    return_raw: bool = False,
    timeout: int = 30,
):
    url = f"{auth.base_url}/projects"

    res = await gd.get_data(
        auth=auth, method="GET", url=url, debug_api=debug_api, timeout=timeout
    )

    if not res.is_success:
        raise gd.BaseError(res=res)

    if return_raw:
        return res

    return [
        AsanaProject._from_json(proj_obj, auth=auth)
        for proj_obj in res.response["data"]
    ]

# %% ../../nbs/asana/01_project.ipynb 8
@patch_to(AsanaProject, cls_method=True)
async def get_by_id(
    cls: AsanaProject,
    auth: aa.AsanaAuth,
    project_id,
    debug_api: bool = False,
    return_raw: bool = False,
    timeout: int = 30,
):
    url = f"{auth.base_url}/projects/{project_id}"

    res = await gd.get_data(
        auth=auth, method="GET", url=url, debug_api=debug_api, timeout=timeout
    )

    if not res.is_success:
        raise gd.BaseError(res=res)

    if return_raw:
        return res

    return cls._from_json(res.response["data"], auth=auth)

# %% ../../nbs/asana/01_project.ipynb 11
@patch_to(AsanaProject)
async def get_tasks(
    self: AsanaProject, debug_api: bool = False, return_raw: bool = False
):
    import domolibrary_extensions.asana.task as at

    auth = self.auth

    url = f"{auth.base_url}/projects/{self.id}/tasks"

    res = await gd.get_data(
        auth=auth,
        method="GET",
        url=url,
        debug_api=debug_api,
        params={
            "opt_fields": "custom_fields, gid, name, assignee, this.assignee.resource_type, this.assignee.email, this.assignee.name, completed, assignee_status, completed_at, created_at, due_on, modified_at, memberships, notes, parent, permalink_url, tags, projects, workspace, this.memberships.section.resource_type, this.memberships.section, this.memberships.project"
        },
    )

    if not res.is_success:
        raise gd.BaseError(res=res)

    if return_raw:
        return res

    self.asana_tasks = [
        at.AsanaTask._from_json(task_obj, auth=auth)
        for task_obj in res.response["data"]
    ]

    return self.asana_tasks



# AUTOGENERATED! DO NOT EDIT! File to edit: ../../nbs/vanilla/classes.ipynb.

# %% auto 0
__all__ = ['Breadcrumb', 'UrlSrcSet', 'Category', 'User', 'Image', 'QuestionStatusEnum', 'Attribute', 'Question', 'Discussion',
           'VanillaForums']

# %% ../../nbs/vanilla/classes.ipynb 7
@dataclass
class Breadcrumb:
    name: str
    url: str

    @classmethod
    def _from_json(cls, json_obj):

        return cls(name=json_obj["name"], url=json_obj["url"])


@dataclass
class UrlSrcSet:

    urls: Dict[str, str]

    @classmethod
    def _from_json(cls, data):

        return cls(urls=data)


@dataclass
class Category:
    base_url: str

    category_id: int
    name: str

    description: str

    parent_category_id: int

    # custom_permissions: bool

    # is_archived: bool

    # urlcode: str
    url: str

    display_as: str

    # icon_url: str

    date_inserted: datetime

    # banner_url: str

    # banner_url_src_set: UrlSrcSet

    count_categories: int
    count_discussions: int
    count_comments: int
    count_all_discussions: int
    count_all_comments: int

    count_followers: int

    # followed: bool

    # featured: bool

    allowed_discussion_types: List[str]

    sort: int = None

    depth: int = None

    # icon_url_src_set: UrlSrcSet

    breadcrumbs: List[Breadcrumb] = None

    @classmethod
    def _from_json(cls, json_obj, base_url):

        category = cls(
            base_url=base_url,
            category_id=json_obj["categoryID"],
            name=json_obj["name"],
            description=json_obj["description"],
            parent_category_id=json_obj["parentCategoryID"],
            url=json_obj["url"],
            display_as=json_obj["displayAs"],
            date_inserted=deuc.convert_str_to_date(json_obj["dateInserted"]),
            # urlcode=json_obj['urlcode'],
            # banner_url=json_obj['bannerUrl'],
            # banner_url_src_set=UrlSrcSet._from_json(json_obj['bannerUrlSrcSet']),
            count_categories=json_obj["countCategories"],
            count_discussions=json_obj["countDiscussions"],
            count_comments=json_obj["countComments"],
            count_all_discussions=json_obj["countAllDiscussions"],
            count_all_comments=json_obj["countAllComments"],
            count_followers=json_obj["countFollowers"],
            allowed_discussion_types=json_obj["allowedDiscussionTypes"],
            depth=json_obj.get("depth"),
            sort=json_obj.get("sort"),
        )

        category.breadcrumbs = [
            Breadcrumb._from_json(breadcrumb) for breadcrumb in json_obj["breadcrumbs"]
        ]

        return category

    @classmethod
    def get_by_id(cls, base_url, category_id, return_raw: bool = False):

        url = f"https://{base_url}/api/v2/categories/{category_id}"

        res = requests.get(url)

        if res.status_code != 200:

            raise Exception(f"unable to retrieve category {category_id}")

        if return_raw:
            return res

        json_obj = res.json()

        return Category._from_json(json_obj, base_url=base_url)

# %% ../../nbs/vanilla/classes.ipynb 9
@dataclass
class User:
    user_id: int
    name: str
    url: str
    photo_url: str
    date_last_active: str
    banned: int
    punished: int
    private: bool
    label: str

    @classmethod
    def _from_json(cls, json_obj):
        return cls(
            user_id=json_obj["userID"],
            name=json_obj["name"],
            url=json_obj["url"],
            photo_url=json_obj["photoUrl"],
            date_last_active=deuc.convert_str_to_date(json_obj["dateLastActive"]),
            banned=json_obj["banned"],
            punished=json_obj["punished"],
            private=json_obj["private"],
            label=json_obj["label"],
        )

# %% ../../nbs/vanilla/classes.ipynb 11
@dataclass
class Image:
    url: str
    urlSrcSet: Dict[str, str]
    alt: str


class QuestionStatusEnum(Enum):
    ANSWERED = auto()
    UNANSWERED = auto()


@dataclass
class Attribute:
    pass


@dataclass
class Question(Attribute):
    status: QuestionStatusEnum = None
    date_accepted: Optional[datetime] = None
    date_answered: datetime = None
    accepted_answers: List[str] = None

    @classmethod
    def _from_json(cls, json_obj):
        """
        Create a Question object from a JSON object.

        Args:
            json_obj (dict): The JSON object.

        Returns:
            Question: The created Question object.
        """
        status = (
            QuestionStatusEnum[json_obj["status"].upper()]
            if json_obj["status"].upper() in QuestionStatusEnum.__members__
            else None
        )
        date_accepted = deuc.convert_str_to_date(json_obj.get("dateAccepted"))

        date_answered = deuc.convert_str_to_date(json_obj.get("dateAnswered"))
        return cls(
            status=status,
            date_accepted=date_accepted,
            date_answered=date_answered,
            accepted_answers=json_obj.get("acceptedAnswers"),
        )

# %% ../../nbs/vanilla/classes.ipynb 13
@dataclass
class Discussion:
    base_url: str
    discussion_id: int
    type: str
    name: str
    body: str
    category_id: int
    date_inserted: datetime
    date_updated: Optional[datetime]
    date_last_comment: str
    insert_user_id: int
    insert_user: User
    closed: bool
    url: str
    canonical_url: str
    status_id: int
    score: Optional[int]

    # count_comments: int
    # count_views: int

    category: Optional[Category] = None
    body_clean: Optional[str] = None
    attributes: Optional[List[Question]] = field(default_factory=dict)

    def _attributes_from_json(self, attributes_dict: dict) -> Attribute:
        """
        Convert a list of attribute dictionaries to a list of Question objects.
        """
        if "question" in attributes_dict:
            self.attributes.update(
                {"question": Question._from_json(attributes_dict["question"])}
            )

        return self.attributes

    @classmethod
    def _from_json(cls, json_obj, base_url):
        discussion = cls(
            base_url=base_url,
            discussion_id=json_obj["discussionID"],
            type=json_obj["type"],
            name=json_obj["name"],
            body=json_obj["body"],
            category_id=json_obj["categoryID"],
            date_inserted=deuc.convert_str_to_date(json_obj.get("dateInserted")),
            date_updated=deuc.convert_str_to_date(json_obj.get("dateUpdated")),
            date_last_comment=deuc.convert_str_to_date(json_obj.get("dateLastComment")),
            insert_user_id=json_obj["insertUserID"],
            insert_user=User._from_json(json_obj["insertUser"]),
            closed=json_obj["closed"],
            # count_comments=json_obj["countComments"],
            # count_views=json_obj["countViews"],
            score=json_obj["score"],
            url=json_obj["url"],
            canonical_url=json_obj["canonicalUrl"],
            status_id=json_obj["statusID"],
        )

        discussion.body_clean = dewb.convert_remove_html_tags(discussion.body)

        if "attributes" in json_obj:
            discussion._attributes_from_json(json_obj["attributes"])

        if "categoryID" in json_obj:
            discussion.category = Category.get_by_id(
                base_url=base_url, category_id=json_obj["categoryID"]
            )

        return discussion

    def to_dict(self) -> dict:
        """
        Convert a Discussion object to a dictionary.

        Returns:
            dict: The dictionary representation of the Discussion object.
        """
        s = {
            "discussion_id": self.discussion_id,
            "discussion_type": self.type,
            "discussion_title": self.name,
            "discussion_body": self.body_clean,
            "category_id": self.category_id,
            "date_inserted": deuc.convert_datetime_to_str(self.date_inserted),
            "date_updated": deuc.convert_datetime_to_str(self.date_updated),
            "date_last_comment": deuc.convert_datetime_to_str(self.date_last_comment),
            "insert_user_id": self.insert_user_id,
            "closed": self.closed,
            # "count_comments": self.count_comments,
            # "count_views": self.count_views,
            "score": self.score,
            "url": self.url,
            "canonical_url": self.canonical_url,
            "status_id": self.status_id,
        }

        if self.category:
            s.update(
                {
                    "category_name": self.category.name,
                    "category_description": self.category.description,
                }
            )

        if self.insert_user:
            s.update(
                {
                    "insert_user_name": self.insert_user.name,
                    "insert_user_profile": self.insert_user.url,
                }
            )

        return deuc.sort_dictionary(s)

# %% ../../nbs/vanilla/classes.ipynb 15
@dataclass
class VanillaForums:
    base_url: str

    discussions: List[Discussion] = None

    def get_all_categories(self, page=1, return_raw: bool = False):
        url = f"https://{self.base_url}/api/v2/categories"
        params = {"page": page}

        res = requests.get(url, params=params)

        if res.status_code != 200:
            raise Exception("unable to retrieve categories")

        if return_raw:
            return res

        categories = res.json()
        self.categories = [
            Category._from_json(category, base_url=self.base_url)
            for category in categories
        ]
        return self.categories

    def get_all_discussions(
        self,
        return_raw: bool = False,
        page: int = 1,
        is_only_unanswered: bool = False,
    ):

        params = {"sort": "-dateInserted", "page": page}

        if is_only_unanswered:
            params.update({"status": "unanswered"})

        url = f"https://{self.base_url}/api/v2/discussions"

        res = requests.get(url, params=params)

        if return_raw:
            return res

        if not res.status_code == 200:
            raise Exception("unable to retrieve discussions")

        discussions = res.json()

        self.discussions = [
            Discussion._from_json(discussion, base_url=self.base_url)
            for discussion in discussions
        ]
        return self.discussions

    def generate_discussion_report(self):
        if not self.discussions:
            self.get_all_discussions()

        return pd.DataFrame([discussion.to_dict() for discussion in self.discussions])



"""supporting functions"""

# AUTOGENERATED! DO NOT EDIT! File to edit: ../../nbs/utils/web.ipynb.

# %% auto 0
__all__ = ['remove_query_params_from_url', 'generate_filename_from_url', 'extract_links']

# %% ../../nbs/utils/web.ipynb 3
from typing import List, Callable, Union
from bs4 import BeautifulSoup
import domolibrary_extensions.utils.files as defi

from urllib.parse import urljoin, urlparse

# %% ../../nbs/utils/web.ipynb 4
from domolibrary_extensions.utils.files import (
    convert_html_to_markdown,
    convert_remove_html_tags,
)

# %% ../../nbs/utils/web.ipynb 5
def remove_query_params_from_url(url):
    u = urlparse(url)
    return urljoin(url, urlparse(url).path)

# %% ../../nbs/utils/web.ipynb 7
def generate_filename_from_url(
    url,
    download_folder=None,
    use_separator="/",
    max_length=None,
    file_name=None,
    suffix=None,
) -> str:
    parsed_url = urlparse(remove_query_params_from_url(url))

    file_path = [str for str in parsed_url[2].split("/") if str][:max_length]

    if download_folder:
        file_path[0:0] = [download_folder]

    if file_name:
        file_path.append(file_name)

    if suffix:
        file_path[-1] = defi.change_file_suffix(file_path[-1], suffix)

    file_path = use_separator.join(file_path)

    return file_path

# %% ../../nbs/utils/web.ipynb 9
def extract_links(
    soup: BeautifulSoup,
    base_url: str = None,
    custom_link_extractor_fn: Callable = None,  # can add custom function for handling URLs
) -> List[Union[str, dict]]:
    """returns a list of urls"""

    links_ls = []

    for link in soup.findAll("a"):
        if not link.has_attr("href"):
            continue

        url = link["href"]

        if custom_link_extractor_fn:
            url = custom_link_extractor_fn(url, base_url)

        if not url:
            continue

        if url.startswith("/") and base_url:
            url = urljoin(base_url, url)

        if base_url and not url.startswith(base_url):
            continue

        if not url or url in links_ls:
            continue

        links_ls.append(url)

    return list(set(links_ls))



# AUTOGENERATED! DO NOT EDIT! File to edit: ../../nbs/scraper/Scraper.ipynb.

# %% ../../nbs/scraper/Scraper.ipynb 2
from __future__ import annotations

import os
import re

from dataclasses import dataclass, field
from typing import List, Callable, Tuple, Any


from functools import partial
from bs4 import BeautifulSoup
from concurrent.futures import ThreadPoolExecutor, as_completed

import domolibrary_extensions.scraper.DriverGenerator as desd
import domolibrary_extensions.scraper.routes as desr
import domolibrary_extensions.utils.utils as deut

from nbdev.showdoc import patch_to

# %% auto 0
__all__ = ['ScrapeConfig', 'is_url_match_pattern', 'ScrapeTask_NoDriverProvided', 'ScrapeTask_NoBaseUrl',
           'ScrapeTask_NoScrapeCrawler', 'ScrapeTask', 'Scrape_Factory_NoConfigMatch', 'ScrapeFactory', 'ScrapeCrawler']

# %% ../../nbs/scraper/Scraper.ipynb 3
from domolibrary_extensions.utils.web import (
    extract_links,
    generate_filename_from_url,
    remove_query_params_from_url,
)

# %% ../../nbs/scraper/Scraper.ipynb 9
@dataclass
class ScrapeConfig:
    """class for collating data about how to scrape a page
    pass a list of scrape_config into ScrapeFactory
    """

    id: re.Pattern  # pattern for matching URLs to appropriate config

    search_criteria_tuple: Tuple = None  # from selenium.webdriver.common.by import By

    content_extractor_fn: Callable = None  # function for the subset of HTML to extract

    link_extractor_fn: Callable = partial(
        extract_links
    )  # function for extracting links from soup

    generate_filename_fn: Callable = (
        None  # fuction for generating filename given a url and dowload_folder
    )

    pagesource_fn: Callable = None

    def __id__(self, other):
        return self.id == other.id

# %% ../../nbs/scraper/Scraper.ipynb 13
def is_url_match_pattern(url, scrape_config_pattern: str):
    """test if a url matches a scrape_config_pattern"""

    pattern = re.compile(scrape_config_pattern)

    match_pattern = pattern.match(url)

    if not match_pattern:
        return False

    return True

# %% ../../nbs/scraper/Scraper.ipynb 15
class ScrapeTask_NoDriverProvided(Exception):
    def __init__(self):
        super().__init__("Driver not provided")


class ScrapeTask_NoBaseUrl(Exception):
    def __init__(self):
        super().__init__(
            "no base_url provided to scrape_task. must run as is_test = True"
        )


class ScrapeTask_NoScrapeCrawler(Exception):
    def __init__(self):
        super().__init__(
            "no scrape_crawler provided.  must run scrape_page as is_test = True"
        )


@dataclass
class ScrapeTask:
    """scrape_factory matches a URL to a scrape_config and returns a scrape_task"""

    url: str  # will be assigned after retrieved from scrape_factory

    base_url: str = None
    download_folder: str = "./SCRAPE"

    generate_filename_fn: Callable = None
    content_extractor_fn: Callable = None
    link_extractor_fn: Callable = None

    pagesource_fn: Callable = None

    max_sleep_time: int = 10
    scrape_config: ScrapeConfig = None

    scrape_config: ScrapeConfig = None
    scrape_crawler: Any = None  # will be optional parent threadpool manager.

    pagesource: str = None

    soup: BeautifulSoup = None

    def __post_init__(self):
        self.generate_filename_fn = (
            self.generate_filename_fn
            or (self.scrape_config and self.scrape_config.generate_filename_fn)
            or partial(generate_filename_from_url, url=self.url, use_separator="/")
        )

        self.pagesource_fn = (
            self.pagesource_fn
            or (self.scrape_config and self.scrape_config.pagesource_fn)
            or partial(
                desr.get_pagesource,
                search_criteria_tuple=(
                    self.scrape_config and self.scrape_config.search_criteria_tuple
                ),
            )
        )

        self.content_extractor_fn = self.content_extractor_fn or (
            self.scrape_config and self.scrape_config.content_extractor_fn
        )

        self.link_extractor_fn = (
            self.link_extractor_fn
            or (self.scrape_config and self.scrape_config.link_extractor_fn)
            or extract_links
        )

    @classmethod
    def _from_factory(
        cls,
        url: str,
        base_url: str,
        scrape_config: ScrapeConfig,
        scrape_crawler=None,
        download_folder: str = "./SCRAPE",
        generate_filename_fn: Callable = None,
    ):
        return cls(
            url=url,
            base_url=base_url,
            download_folder=download_folder,
            scrape_config=scrape_config,
            scrape_crawler=scrape_crawler,
            generate_filename_fn=generate_filename_fn,
        )

    def _download_content(self, file_name, content):

        dir_name = os.path.dirname(file_name)

        if dir_name[-1] != "/":
            dir_name += "/"

        deut.upsert_folder(dir_name)

        with open(file_name, "w", encoding="utf-8") as f:
            f.write(str(content))

        return True

    def _update_parent(self):
        """if config is part of a scrape_crawler / managed threadpool will update crawler's urls_to_visit list."""

        scrape_crawler = self.scrape_crawler

        if not scrape_crawler:
            return False

        scrape_crawler._add_url_to_visited(self.url)

        if self.soup:
            urls_to_visit = self.link_extractor_fn(
                soup=self.soup, base_url=self.base_url
            )

            print(f"{len(urls_to_visit)} exracted from soup")

            [scrape_crawler._add_url_to_visit(url) for url in urls_to_visit]

            print(self.scrape_crawler.urls_to_visit, self.base_url)

        return urls_to_visit

# %% ../../nbs/scraper/Scraper.ipynb 16
@patch_to(ScrapeTask)
def execute(
    self: ScrapeTask,
    driver=None,
    debug_prn: bool = False,
    is_suppress_errors: bool = False,
    is_test: bool = False,
    download_folder: str = None,
):
    """handles executing the scrape_task
    1. get pagesource
    2. download index
    3. download content
    4. update crawler
    - get_links from pagesource
    """

    try:
        self.pagesource = self.pagesource_fn(
            driver=driver,
            url=self.url,
            max_sleep_time=self.max_sleep_time,
            return_entire_page=True,
            return_soup=False,
            debug_prn=debug_prn,
        )

        download_folder = download_folder or self.download_folder

        self.soup = BeautifulSoup(self.pagesource, "lxml")

        # download index
        file_name = self.generate_filename_fn(
            url=self.url, download_folder=download_folder, file_name="index.html"
        )

        self._download_content(file_name=file_name, content=self.soup)

        # download content
        if self.content_extractor_fn:
            content = self.content_extractor_fn(self.soup)
            content_name = "content.html"

            if isinstance(content, tuple):
                content_name = content[1]
                content = content[0]

            content_name = self.generate_filename_fn(
                download_folder=download_folder, file_name=content_name, url=self.url
            )

            self._download_content(
                file_name=content_name,
                content=content,
            )

        print(f"🎉 successfully scraped {self.url} to {file_name}")

        if is_test or not self.scrape_crawler:
            return True

        # update crawler
        # if not self.base_url:
        #     raise ScrapeTask_NoBaseUrl()

        # if not self.scrape_crawler:
        #     raise ScrapeTask_NoScrapeCrawler()

        """if config is part of a scrape_crawler / managed threadpool will update crawler's urls_to_visit list."""
        self._update_parent()

        return True

    except Exception as e:
        if not is_suppress_errors:
            raise (e)
        return f"💀 failed to download {self.url} received errror{e}"

# %% ../../nbs/scraper/Scraper.ipynb 20
class Scrape_Factory_NoConfigMatch(Exception):
    def __init__(self, text):
        super().__init__(
            f"{text} has no pattern match in factory_configs, add an appropriate config or check pattern matches"
        )


@dataclass
class ScrapeFactory:
    """class handles a list of Scrape_Configs and will return the 'correct one' given a URL"""

    factory_configs: List[ScrapeConfig]

    def generate_task_from_url(
        self,
        url,
        download_folder="./SCRAPE",
        base_url=None,
        scrape_crawler=None,
        generate_filename_fn=None,
    ):
        configs = []

        for config in self.factory_configs:
            if is_url_match_pattern(url=url, scrape_config_pattern=config.id):
                configs.append(config)

        if not configs:
            raise Scrape_Factory_NoConfigMatch(text=url)

        return [
            ScrapeTask._from_factory(
                scrape_config=config,
                url=url,
                base_url=base_url,
                download_folder=download_folder,
                scrape_crawler=scrape_crawler,
                generate_filename_fn=generate_filename_fn,
            )
            for config in configs
        ]

# %% ../../nbs/scraper/Scraper.ipynb 22
@dataclass
class ScrapeCrawler:
    """threadpool manager for crawling through a list of urls"""

    base_url: str
    download_folder: str
    visited_urls: set
    urls_to_visit: set

    executor: ThreadPoolExecutor = field(repr=False)
    scrape_factory: ScrapeFactory = field(repr=False)  # list of expected scrape_configs
    driver_generator: desd.DriverGenerator = field(repr=False)

    # def __post_init__(self):

    @classmethod
    def init(
        cls,
        urls_to_visit: List[str],
        base_url: str,
        scrape_factory: ScrapeFactory,
        driver_generator: desd.DriverGenerator = None,
        download_folder: str = "./SCRAPE/",
        driver_path=None,
        max_workers=5,
        urls_visited: List[str] = None,
    ):
        scrape_crawler = cls(
            executor=ThreadPoolExecutor(max_workers=max_workers),
            scrape_factory=scrape_factory,
            base_url=base_url,
            download_folder=download_folder,
            visited_urls=set(),
            urls_to_visit=set(),
            driver_generator=driver_generator
            or desd.DriverGenerator(driver_path=driver_path),
        )

        if urls_visited:
            [scrape_crawler._add_url_to_visited(url) for url in urls_visited]

        if urls_to_visit:
            [scrape_crawler._add_url_to_visit(url) for url in urls_to_visit]

        return scrape_crawler

    def _add_url_to_visit(self, url, debug_prn: bool = False):
        """adds a URL to the list of URLS to visit after testing that the URL has not already been visited"""

        if url not in self.visited_urls and url not in self.urls_to_visit:
            if debug_prn:
                print(f"adding {url} to to_vist list")

            self.urls_to_visit.add(url)
            return self.urls_to_visit

    def _add_url_to_visited(self, url, debug_prn: bool = False):
        if url not in self.visited_urls:
            if debug_prn:
                print(f"adding {url} to visited list")

            self.visited_urls.add(url)
            return self.visited_urls

    def _quit(self):
        """call when the executor queue is empty"""

        self.executor.shutdown(wait=True)
        return f"Done scraping {len(self.visited_urls)} urls"

    def crawl_urls(
        self: ScrapeCrawler,
        is_test: bool = False,
        debug_prn: bool = False,
        is_headless: bool = True,
    ):
        while self.urls_to_visit:
            url = self.urls_to_visit.pop()
            self.visited_urls.add(url)

            driver = self.driver_generator.get_webdriver(is_headless=is_headless)

            tasks = self.scrape_factory.generate_task_from_url(
                url,
                scrape_crawler=self,
                base_url=self.base_url,
                download_folder=self.download_folder,
            )

            future_to_task = {
                self.executor.submit(task.execute, driver=driver, is_test=is_test): task
                for task in tasks
            }
            for future in as_completed(future_to_task):

                try:
                    future.result()

                except Exception as e:
                    print(e)

        return self._quit()



"""supporting functions"""

# AUTOGENERATED! DO NOT EDIT! File to edit: ../../nbs/utils/utils.ipynb.

# %% auto 0
__all__ = ['printmd']

# %% ../../nbs/utils/utils.ipynb 3
from .convert import *
from .files import *

# %% ../../nbs/utils/utils.ipynb 4
from IPython.display import Markdown, display

# %% ../../nbs/utils/utils.ipynb 5
def printmd(string):
    display(Markdown(string))



# AUTOGENERATED! DO NOT EDIT! File to edit: ../../nbs/asana/00_auth.ipynb.

# %% ../../nbs/asana/00_auth.ipynb 2
from __future__ import annotations

from dataclasses import dataclass, field

import domolibrary_extensions.client as gd

# %% auto 0
__all__ = ['AsanaAuth']

# %% ../../nbs/asana/00_auth.ipynb 4
@dataclass
class AsanaAuth(gd.Auth):
    token: str = field(repr=False)
    workspace_id: str
    instance: str = field(repr=False, default=None)
    base_url = "https://app.asana.com/api/1.0"
    auth_header: dict = field(repr=False, default=None)

    def __post_init__(self):
        self.generate_auth_header()
        self.instance = self.workspace_id

    def get_auth_token(self):
        return self.token

    def generate_auth_header(self):
        self.auth_header = {
            "Authorization": f"Bearer {self.token}",
        }
        return self.auth_header

    async def who_am_i(self):
        url = f"{self.base_url}/users/me"
        res = await gd.get_data(method="get", url=url, auth=self)

        if not res.is_success:
            raise gd.BaseError(res=res)

        return res
